"""A flexible resource for the root of your API resource tree."""

import logging
from collections import defaultdict, namedtuple
from threading import RLock
from typing import Callable, Dict, Optional, OrderedDict

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest, HttpResponseNotModified
from django.utils.translation import gettext_lazy as _
from django.urls import path
from typing_extensions import TypeAlias

from djblets.webapi.errors import DOES_NOT_EXIST
from djblets.webapi.resources.base import WebAPIResource
from djblets.webapi.responses import WebAPIResponseError


logger = logging.getLogger(__name__)


#: A type alias for a mapping of URI templates.
#:
#: Each key is a URI template name, and each value is a URL.
#:
#: Version Added:
#:     3.2
URITemplates: TypeAlias = Dict[str, str]


#: A mapping of API resource instances to URI templates.
#:
#: Version Added:
#:     3.2
_URITemplatesByResourceMap: TypeAlias = Dict[Optional[WebAPIResource],
                                             URITemplates]


class _URITemplatesCache:
    """A least-recently-used cache for URI templates.

    This will map base API URLs to generated dictionaries of URI templates. It
    holds up to a maximum number of cached entries, and will discard the oldest
    as needed once the limit is reached.

    Since cache state is shared between threads, adding and fetching items are
    done behind a thread lock, to avoid issues with multiple threads trying to
    manipulate the cache at the same time.

    Version Added:
        3.2
    """

    ######################
    # Instance variables #
    ######################

    #: The maximum number of URI templates to store in the cache.
    #:
    #: Type:
    #:     int
    max_size: int

    def __init__(
        self,
        max_size: int = 50,
    ) -> None:
        """Initialize the cache.

        Args:
            max_size (int, optional):
                The maximum size of the cache.
        """
        self.max_size = max_size
        self._cache = OrderedDict[str, URITemplates]()
        self._lock = RLock()

    def add(
        self,
        base_href: str,
        uri_templates: URITemplates,
    ) -> None:
        """Add URI templates to the cache.

        If the cache is full, the last item will be discarded.

        Args:
            base_href (str):
                The base API URL mapped to the URI templates.

            uri_templates (dict):
                The dictionary of generated URI templates.
        """
        with self._lock:
            cache = self._cache

            if len(cache) >= self.max_size:
                cache.popitem(last=False)

            cache[base_href] = uri_templates

    def get(
        self,
        base_href: str,
        *,
        build_func: Optional[Callable[[str], URITemplates]] = None,
    ) -> URITemplates:
        """Return URI templates from the cache.

        If the URI templates are in the cache, they'll be returned and then
        marked as most-recently used, to avoid falling out of cache too soon.

        Args:
            base_href (str):
                The base API URL mapped to the URI templates.

            build_func (callable, optional):
                Optional function to call to build the results if missing.

                This will be built while a lock is held.

        Returns:
            dict:
            The URI templates found in the cache. This may be
            newly-constructed.

        Raises:
            KeyError:
                No cached content exists for the given ``base_href``.
        """
        with self._lock:
            cache = self._cache

            try:
                uri_templates = cache[base_href]
                cache.move_to_end(base_href)
            except KeyError:
                if build_func is None:
                    raise

                uri_templates = build_func(base_href)
                self.add(base_href, uri_templates)

            return uri_templates

    def clear(self) -> None:
        """Clear the URI template cache."""
        with self._lock:
            self._cache.clear()


class RootResource(WebAPIResource):
    """The root of a resource tree.

    This is meant to be instantiated with a list of immediate child
    resources. The result of :py:meth:`get_url_patterns` should be included in
    a project's :file:`urls.py`.
    """

    name = 'root'
    singleton = True

    #: A resource entry returned from :py:meth:`RootResource.walk_resources`.
    #:
    #: Attributes:
    #:      name (str):
    #:          The name of current resource being explored.
    #:
    #       list_href (str):
    #:          The list_href associated with this resource.
    #:
    #       resource (djblets.webapi.resources.base.WebAPIResource):
    #:          The resource object being explored.
    #:
    #       is_list (bool):
    #:          Whether or not the resource is a list resource.
    #:
    #:      uri_template_name (str)
    #:          The name of the resource to use in the URI templates list.
    ResourceEntry = namedtuple('ResourceEntry',
                               ('name',
                                'list_href',
                                'resource',
                                'is_list',
                                'uri_template_name'))

    def __init__(self, child_resources=[], include_uri_templates=True):
        super(RootResource, self).__init__()
        self.list_child_resources = child_resources

        self._cached_uri_templates = _URITemplatesCache()
        self._registered_uri_templates: _URITemplatesByResourceMap = {}
        self._include_uri_templates = include_uri_templates

    def get_etag(self, request, obj, *args, **kwargs):
        return self.encode_etag(request, repr(obj))

    def get(self, request, *args, **kwargs):
        """Retrieve the list of top-level resources and URL templates."""
        data = self.serialize_root(request, *args, **kwargs)
        etag = self.get_etag(request, data)

        if self.are_cache_headers_current(request, etag=etag):
            return HttpResponseNotModified()

        return 200, data, {
            'ETag': etag,
        }

    def serialize_root(self, request, *args, **kwargs):
        """Serialize the contents of the root resource.

        By default, this just provides links and URI templates. Subclasses
        can override this to provide additional data, or to otherwise
        change the structure of the root resource.
        """
        data = {
            'links': self.get_links(self.list_child_resources,
                                    request=request, *args, **kwargs),
        }

        if self._include_uri_templates:
            data['uri_templates'] = self.get_uri_templates(request, *args,
                                                           **kwargs)

        return data

    def get_uri_templates(
        self,
        request: HttpRequest,
        *args,
        **kwargs,
    ) -> URITemplates:
        """Return all URI templates in the resource tree.

        REST APIs can be very chatty if a client wants to be well-behaved
        and crawl the resource tree asking for the links, instead of
        hard-coding the paths. The benefit is that they can keep from
        breaking when paths change. The downside is that it can take many
        HTTP requests to get the right resource.

        This list of all URI templates allows clients who know the resource
        name and the data they care about to simply plug them into the
        URI template instead of trying to crawl over the whole tree. This
        can make things far more efficient.

        Args:
            request (django.http.HttpRequest):
                The GET request for the Root resource.

            *args (tuple, unused):
                Additional unused arguments.

            **kwargs (dict, unused):
                Additional unused keyword arguments.

        Returns:
            dict:
            A mapping of resources to their URI templates.
        """
        base_href = request.build_absolute_uri()

        return self._cached_uri_templates.get(
            base_href,
            build_func=self.build_uri_templates)

    def build_uri_templates(
        self,
        base_href: str,
    ) -> URITemplates:
        """Build the URI templates to include in the payload.

        The resource will cache these automatically in local memory, so that
        repeated calls will often be unnecessary. This allows for things like
        per-organization APIs that each have their own cache.

        Caches are automatically invalidated as needed.

        Version Added:
            3.2

        Args:
            base_href (str):
                The base URL for the API endpoint.

        Returns:
            dict:
            A dictionary mapping URI template keys to template values.
        """
        templates = defaultdict(list)
        unassigned_templates = self._registered_uri_templates.get(
            None, {})

        for uri_template_name, href in unassigned_templates.items():
            templates[uri_template_name].append(href)

        for entry in self.walk_resources(self, base_href):
            uri_template_name = entry.uri_template_name

            if uri_template_name is None:
                continue

            templates[uri_template_name].append(entry.list_href)

            if entry.is_list:
                list_templates = self._registered_uri_templates.get(
                    entry.resource, {})

                for uri_template_name, href in list_templates.items():
                    templates[uri_template_name].append(
                        '%s%s' % (entry.list_href, href))

        final_templates = {}

        for uri_template_name, uri_template in templates.items():
            if len(uri_template) > 1:
                if settings.DEBUG or getattr(settings, 'PRODUCTION', True):
                    logger.error(
                        'More than one URI template was mapped to '
                        'the "%s" name: %s. Only the first one will be '
                        'included in the URI templates list. To '
                        'include the other URI templates, they must '
                        'be mapped to a unique name by setting each '
                        'resource\'s uri_template_name property.'
                        % (uri_template_name, ', '.join(uri_template)))
                else:
                    raise ImproperlyConfigured(
                        _('More than one URI template was mapped to '
                          'the "%(name)s" name: %(templates)s. Each URI '
                          'template must be mapped to a unique URI '
                          'template name in order to be included in the '
                          'URI templates list. This can be set through '
                          'the uri_template_name property.')
                        % {
                            'name': uri_template_name,
                            'templates': ', '.join(uri_template)
                        })

            final_templates[uri_template_name] = uri_template[0]

        return final_templates

    @classmethod
    def walk_resources(cls, resource, list_href):
        """Yield all URI endpoints associated with a specified resource.

        Args:
            resource (djblets.webapi.resources.WebAPIResource):
                The starting point for searching the resource tree.

            list_href (unicode):
                The path to the list resource, relative to the WebAPIResource
                provided. Used as a component of the URL in the API.

        Yields:
            RootResource.ResourceEntry:
            Resource entries for all sub-resources.
        """
        yield cls.ResourceEntry(
            name=resource.name_plural,
            list_href=list_href,
            resource=resource,
            is_list=True,
            uri_template_name=resource.uri_template_name_plural)

        for child in resource.list_child_resources:
            child_href = '%s%s/' % (list_href, child.uri_name)

            for entry in cls.walk_resources(child, child_href):
                yield entry

        if resource.uri_object_key:
            object_href = '%s{%s}/' % (list_href, resource.uri_object_key)

            yield cls.ResourceEntry(
                name=resource.name,
                list_href=object_href,
                resource=resource,
                is_list=False,
                uri_template_name=resource.uri_template_name)

            for child in resource.item_child_resources:
                child_href = '%s%s/' % (object_href, child.uri_name)

                for entry in cls.walk_resources(child, child_href):
                    yield entry

    def api_404_handler(self, request, api_format=None, *args, **kwargs):
        """Default handler at the end of the URL patterns.

        This returns an API 404, instead of a normal django 404.
        """
        return WebAPIResponseError(request,
                                   err=DOES_NOT_EXIST,
                                   api_format=api_format)

    def get_url_patterns(self):
        """Return the Django URL patterns for this object and its children.

        This returns the same list as
        :py:meth:`WebAPIResource.get_url_patterns`, but also introduces a
        generic catch-all 404 handler which returns API errors instead of HTML.
        """
        urlpatterns = super(RootResource, self).get_url_patterns()
        urlpatterns += [
            path('<str>', self.api_404_handler),
        ]

        return urlpatterns

    def register_uri_template(
        self,
        name: str,
        relative_path: str,
        relative_resource: Optional[WebAPIResource] = None,
    ) -> None:
        """Register the specified resource for URI template serialization.

        This adds the specified name and relative resource to the Root
        Resource's URI templates.

        Args:
            name (str):
                The name of the associated resource being added to templates.

            relative_path (str):
                The path of the API resource relative to its parent resources.

            relative_resource (djblets.webapi.resources.base.WebAPIResource,
                               optional):
                The resource instance associated with this URI template.
        """
        templates = self._registered_uri_templates.setdefault(
            relative_resource, {})

        if name in templates:
            logger.debug(
                'The %s resource is already mapped to the following '
                'URI template: %s. This will be overwritten by the new URI '
                'template: %s.'
                % (name, templates[name], relative_path))

        # Allow the URI template to be overwritten if it already exists.
        templates[name] = relative_path

        # Clear the cache so that new lookups can detect newly added templates.
        self._cached_uri_templates.clear()

    def unregister_uri_template(
        self,
        name: str,
        relative_resource: Optional[WebAPIResource] = None,
    ) -> None:
        """Unregister the specified resource for URI template serialization.

        This removes the specified name and relative resource to the Root
        Resource's URI templates.

        Args:
            name (str):
                The name of the resource being removed from templates.

            relative_resource (djblets.webapi.resources.base.WebAPIResource,
                               optional):
                The resource instance associated with this URI template.
        """
        try:
            del self._registered_uri_templates[relative_resource][name]
        except KeyError:
            return

        # Clear the cache so that new lookups won't include the old URI
        # templates.
        self._cached_uri_templates.clear()



"""Support for handling deprecations.

The version-specific objects in this module are not considered stable between
releases, and may be removed at any point.

Version Changed:
    4.0:
    Projects should now use :pypi:`housekeeping` to manage their own
    deprecations, instead of using this module.
"""

from __future__ import annotations

from typing import Any, Callable, Optional, Type

from housekeeping import (
    BasePendingRemovalWarning,
    BaseRemovedInWarning,
    deprecated_arg_value as djblets_deprecated_arg_value,
    deprecate_non_keyword_only_args as djblets_deprecate_non_keyword_only_args,
    func_moved)
from housekeeping.functions import _FuncT
from housekeeping.helpers import LazyObject


class BaseRemovedInProductVersionWarning(BaseRemovedInWarning):
    """Base class for product deprecation warnings.

    This can be used by consumers of Djblets to provide similar deprecation
    functionality, making use of deprecation utility functions.

    Deprecated:
        4.0:
        Consumers should subclass :py:class:`housekeeping.BaseRemovedInWarning`
        instead. This class will be removed in Djblets 6.

    Version Added:
        3.2
    """


class PendingRemovalInDjbletsWarning(BasePendingRemovalWarning):
    """Pending deprecation for code in Djblets.

    Version Added:
        4.0
    """

    # Don't warn that we're subclassing BaseRemovedInProductVersionWarning,
    # which is needed to continue being able to catch warnings in a generic
    # fashion. This can be removed in Djblets 6.
    housekeeping_skip_warning = True

    product = 'Djblets'


class BaseRemovedInDjbletsVersionWarning(BaseRemovedInProductVersionWarning):
    """Base class for a Djblets deprecation warning.

    All version-specific deprecation warnings inherit from this, allowing
    callers to check for Djblets deprecations without being tied to a specific
    version.
    """

    product = 'Djblets'


class RemovedInDjblets60Warning(BaseRemovedInDjbletsVersionWarning):
    """Deprecations for features scheduled for removal in Djblets 6.0.

    Note that this class will itself be removed in Djblets 6.0. If you need to
    check against Djblets deprecation warnings, please see
    :py:class:`BaseRemovedInDjbletsVersionWarning`.
    """

    version = '6.0'


class RemovedInDjblets70Warning(BaseRemovedInDjbletsVersionWarning):
    """Deprecations for features scheduled for removal in Djblets 7.0.

    Note that this class will itself be removed in Djblets 7.0. If you need to
    check against Djblets deprecation warnings, please see
    :py:class:`BaseRemovedInDjbletsVersionWarning`.
    """

    version = '7.0'


#: An alias for the next release of Djblets where features would be removed.
RemovedInNextDjbletsVersionWarning = RemovedInDjblets60Warning


@func_moved(RemovedInDjblets60Warning,
            new_func=djblets_deprecated_arg_value)
def deprecated_arg_value(
    owner_name: str,
    value: Any,
    old_arg_name: str,
    new_arg_name: Optional[str] = None,
    warning_cls: Type[BaseRemovedInWarning] = BaseRemovedInWarning,
) -> LazyObject:
    """Wrap a value in a lazy object to warn when used.

    Deprecated:
        4.0:
        Consumers should use :py:func:`housekeeping.deprecated_arg_value`
        instead.

    Version Changed:
        3.2:
        Specifying a :py:class:`BaseRemovedInProductVersionWarning` subclass
        will now result in product/version information in the warning message.

    Args:
        owner_name (str):
            The name of the owner of this argument.

        value (object):
            The argument value.

        old_arg_name (str):
            The name of the argument that was deprecated.

        new_arg_name (str, optional):
            The optional name of the argument to use in the deprecated
            argument's place, if one is available.

        warning_cls (type, optional):
            The class to use for the warning. This should be
            :py:exc:`DeprecationWarning`, :py:exc:`PendingDeprecationWarning`,
            or a subclass of one.

            If a :py:class:`BaseRemovedInProductVersionWarning` subclass is
            provided, the error message will include warning information.
            This is recommended and may be required in a future version.

    Returns:
        django.utils.functional.SimpleLazyObject:
        The value wrapped in a lazy object. The first time it is accessed,
        a warning will be emitted.
    """
    return djblets_deprecated_arg_value(warning_cls=warning_cls,
                                        owner_name=owner_name,
                                        value=value,
                                        old_name=old_arg_name,
                                        new_name=new_arg_name)


@func_moved(RemovedInDjblets60Warning,
            new_func=djblets_deprecate_non_keyword_only_args)
def deprecate_non_keyword_only_args(
    warning_cls: Type[BaseRemovedInProductVersionWarning],
    message: Optional[str] = None,
) -> Callable[[_FuncT], _FuncT]:
    """Deprecate calls passing keyword-only arguments as positional arguments.

    This decorator allows code transitioning to keyword-only arguments to
    continue working when passing values as positional arguments.

    Upon the first call, it will record information about the signature of the
    function and then compare that to any future calls. If any positional
    argument values are passed to keyword-only arguments, the arguments will
    be rewritten to work correctly, and a deprecation warning will be emitted.

    Deprecated:
        4.0:
        Consumers should use
        :py:func:`housekeeping.deprecate_non_keyword_only_args`
        instead.

    Version Added:
        3.2

    Args:
        warning_cls (type):
            The specific deprecation warning class to use. This must be a
            subclass of :py:exc:`DeprecationWarning`.

        message (str, optional):
            An optional message to use instead of the default.

    Returns:
        callable:
        The function decorator.

    Raises:
        AssertionError:
            The function being called does not provide keyword-only arguments.
    """
    return djblets_deprecate_non_keyword_only_args(warning_cls=warning_cls,
                                                   message=message)



"""Error classes and codes for WebAPI."""

from __future__ import annotations

from typing import Callable, Dict, List, Optional, TYPE_CHECKING, Union

from housekeeping import deprecate_non_keyword_only_args

from djblets.deprecation import RemovedInDjblets60Warning

if TYPE_CHECKING:
    from django.http import HttpRequest
    from djblets.webapi.responses import WebAPIResponseHeaders
    from typing_extensions import Self, TypeAlias

    #: A mapping of HTTP headers to values, or a callable returning headers.
    #:
    #: Version Added:
    #:     4.0
    _HTTPHeadersOrCallable: TypeAlias = Union[
        WebAPIResponseHeaders,
        Callable[[HttpRequest], WebAPIResponseHeaders]
    ]


class WebAPIError:
    """An API error, containing an error code and human readable message.

    Version Changed:
        4.0:
        * Added Python type hints.
        * Added :py:attr:`detail`, :py:attr:`error_type`, and
          :py:attr:`error_subtype`, and :py:attr:`trace_id` attributes.
    """

    ######################
    # Instance variables #
    ######################

    #: The numeric code for the error.
    #:
    #: This may be used along with :py:attr:`error_type`, depending on the
    #: consumer's API requirements. This was the original error type indicator
    #: prior to Djblets 4.
    #:
    #: Type:
    #:     int
    code: int

    #: A message used to convey additional detail on this error.
    #:
    #: Version Added:
    #:     4.0
    detail: Optional[str]

    #: A string indicating the subtype of the error.
    #:
    #: This can provide additional context beyond :py:attr:`code` or
    #: :py:attr:`error_type`.
    #:
    #: Version Added:
    #:     4.0
    error_subtype: Optional[str]

    #: A string indicating the type of the error.
    #:
    #: This is a string equivalent to ``code``, and is used to allow callers
    #: to differentiate between errors. It may be accompanied by a
    #: :py:attr:`error_subtype``.
    #:
    #: Version Added:
    #:     4.0
    error_type: Optional[str]

    #: Extra HTTP headers included in the error's HTTP response.
    #:
    #: This may also be a callable that returns HTTP headers, given an HTTP
    #: request.
    #:
    #: Type:
    #:     dict or callable
    headers: _HTTPHeadersOrCallable

    #: The HTTP status code of the error.
    #:
    #: Type:
    #:     int
    http_status: int

    #: The error message shown to the caller.
    #:
    #: Type:
    #:     str
    msg: str

    #: An ID shown in the error that's associated with logs.
    #:
    #: This can be supplied using :py:meth:`with_overrides` in order to
    #: associate the error with logged output containing more detail.
    #:
    #: Version Added:
    #:     4.0
    trace_id: Optional[str]

    @deprecate_non_keyword_only_args(RemovedInDjblets60Warning)
    def __init__(
        self,
        code: int,
        msg: str,
        *,
        http_status: int = 400,
        headers: _HTTPHeadersOrCallable = {},
        error_type: Optional[str] = None,
        error_subtype: Optional[str] = None,
        detail: Optional[str] = None,
        trace_id: Optional[str] = None,
    ) -> None:
        """Initialize the error.

        Version Changed:
            4.0:
            * Added ``error_type``, ``error_subtype``, ``detail``, and
              ``trace_id`` attributes.
            * ``http_status`` and ``headers`` must be provided as a keyword.
              Passing as positional arguments is deprecated and will be removed
              in Djblets 6.

        Args:
            code (int):
                The numeric code for the error.

            msg (str):
                The error message shown to the caller.

            http_status (int, optional):
                The HTTP status code for the error.

                This defaults to :http:`400`.

            headers (dict or callable, optional):
                Extra HTTP headers to include in the error's HTTP response.

                This may also be a callable that returns HTTP headers, given
                an HTTP request.

            error_type (str, optional):
                A string indicating the type of the error.

                This is a string equivalent to ``code``, and is used to allow
                callers to differentiate between errors. It may be accompanied
                by ``error_subtype``.

                Third-party errors (such as those provided by extensions)
                should be in the form of :samp:`x-{vendor}-{type}`.

                Version Added:
                    4.0

            error_subtype (str, optional):
                A string indicating the subtype of the error.

                This can provide additional context beyond :py:attr:`code` or
                :py:attr:`error_type`.

                Third-party errors (such as those provided by extensions)
                should be in the form of :samp:`x-{vendor}-{type}`.

                Version Added:
                    4.0

            detail (str, optional):
                An message used to convey additional details on this error.

                Version Added:
                    4.0

            trace_id (str, optional):
                An ID shown in the error that's associated with logs.

                This can be supplied using :py:meth:`with_overrides` in order
                to associate the error with logged output containing more
                detail.

                Version Added:
                    4.0
        """
        self.code = code
        self.msg = msg
        self.http_status = http_status
        self.headers = headers
        self.error_type = error_type
        self.error_subtype = error_subtype
        self.detail = detail
        self.trace_id = trace_id

    def __repr__(self) -> str:
        """Return a string representation of the error.

        Returns:
            str:
            The string representation.
        """
        return '<API Error %d, HTTP %d: %s>' % (self.code, self.http_status,
                                                self.msg)

    @deprecate_non_keyword_only_args(RemovedInDjblets60Warning)
    def with_overrides(
        self,
        msg: Optional[str] = None,
        *,
        headers: Optional[_HTTPHeadersOrCallable] = None,
        error_subtype: Optional[str] = None,
        detail: Optional[str] = None,
        trace_id: Optional[str] = None,
    ) -> Self:
        """Return an error with overridden values.

        The returned error will be based on this error, but with a custom
        message or headers.

        Version Changed:
            4.0:
            * Added ``error_subtype``, ``detail``, and ``trace_id`` attributes.
            * ``headers`` must be provided as a keyword. Passing as a
              positional argument is deprecated and will be removed in
              Djblets 6.

        Args:
            msg (str, optional):
                The custom error message.

            headers (dict or callable, optional):
                Custom HTTP headers to include in the error's HTTP response.

                This may also be a callable that returns HTTP headers, given
                an HTTP request.

                These will override the default HTTP headers for the error.

            error_subtype (str, optional):
                A string indicating the subtype of the error.

                This can provide additional context beyond :py:attr:`code` or
                :py:attr:`error_type`.

                Third-party errors (such as those provided by extensions)
                should be in the form of :samp:`x-{vendor}-{type}`.

                Version Added:
                    4.0

            detail (str, optional):
                An message used to convey additional details on this error.

                Version Added:
                    4.0

            trace_id (str, optional):
                An ID shown in the error that's associated with logs.

                This can be supplied using :py:meth:`with_overrides` in order
                to associate the error with logged output containing more
                detail.

                Version Added:
                    4.0

        Returns:
            WebAPIError:
            The new error instance.
        """
        if headers is None:
            headers = self.headers

        return type(self)(code=self.code,
                          msg=msg or self.msg,
                          http_status=self.http_status,
                          headers=headers,
                          error_type=self.error_type,
                          error_subtype=error_subtype or self.error_subtype,
                          detail=detail or self.detail,
                          trace_id=trace_id or self.trace_id)

    def with_message(
        self,
        msg: str,
    ) -> Self:
        """Return an error with an overridden error message.

        The returned error will be based on this error, but with a custom
        message.

        Args:
            msg (str):
                The custom error message.

        Returns:
            WebAPIError:
            The new error instance.
        """
        return self.with_overrides(msg)


class WebAPITokenGenerationError(Exception):
    """An error generating a Web API token."""


def _get_auth_headers(
    request: HttpRequest,
) -> WebAPIResponseHeaders:
    """Return HTTP headers for an authentication error.

    The result will include any headers from the authentication backend,
    along with a :mailheader:`WWW-Authenticate` header containing a
    comma-separated list of schemes set by auth backends (see
    :py:attr:`WebAPIAuthBackend.www_auth_scheme
    <djblets.webapi.auth.backends.base.WebAPIAuthBackend.www_auth_scheme>`.

    Args:
        request (django.http.HttpRequest):
            The HTTP request from the client.

    Returns:
        dict:
        The mapping of HTTP authentication headers to values.
    """
    from djblets.webapi.auth.backends import get_auth_backends

    headers: Dict[str, str] = {}
    www_auth_schemes: List[str] = []

    for auth_backend_cls in get_auth_backends():
        auth_backend = auth_backend_cls()

        if auth_backend.www_auth_scheme:
            www_auth_schemes.append(auth_backend.www_auth_scheme)

        headers.update(auth_backend.get_auth_headers(request))

    if www_auth_schemes:
        headers['WWW-Authenticate'] = ', '.join(www_auth_schemes)

    return headers


#: An error indicating no error occurred.
#:
#: This should never be encountered.
NO_ERROR = WebAPIError(
    code=0,
    msg='If you see this, yell at the developers',
    error_type='no-error')


#: An error indicating the web service is not configured.
SERVICE_NOT_CONFIGURED = WebAPIError(
    code=1,
    msg='The web service has not yet been configured',
    error_type='server-not-configured',
    http_status=503)


#: An error indicating a requested resource does not exist.
DOES_NOT_EXIST = WebAPIError(
    code=100,
    msg='Object does not exist',
    error_type='resource-does-not-exist',
    http_status=404)


#: An error indicating the user doesn't have permission for a resource.
#:
#: The user may not have read access or write access to the resource.
PERMISSION_DENIED = WebAPIError(
    code=101,
    msg="You don't have permission for this",
    error_type='resource-permission-denied',
    http_status=403)


#: An error indicating a resource doesn't support an attribute for the request.
#:
#: This should be used when performing a request to a resource that's dependent
#: on some attribute (such as a capability flag, setting, or associated
#: resource) on the requested resource being available or set to an expected
#: value, or when the request is missing the attribute.
#:
#: This response payload should contain a top-level ``reason`` field with an
#: error message.
#:
#: For invalid request fields, see :py:data:`INVALID_FORM_DATA`.
INVALID_ATTRIBUTE = WebAPIError(
    code=102,
    msg='Invalid attribute',
    error_type='resource-attribute-invalid',
    http_status=400)


#: An error indicating that the user is not logged in.
#:
#: This can be used to require authenticating to the API before making the
#: request.
NOT_LOGGED_IN = WebAPIError(
    code=103,
    msg='You are not logged in',
    error_type='auth-not-logged-in',
    http_status=401,
    headers=_get_auth_headers)


#: An error indicating that authentication failed.
LOGIN_FAILED = WebAPIError(
    code=104,
    msg='The username or password was not correct',
    error_type='auth-login-failed',
    http_status=401,
    headers=_get_auth_headers)


#: An error indicating data submitted in the request is invalid.
#:
#: Callers should set ``fields`` at the top of the payload to a dictionary
#: mapping field names to lists of error mesages.
INVALID_FORM_DATA = WebAPIError(
    code=105,
    msg='One or more fields had errors',
    error_type='request-field-error',
    http_status=400)


#: An error indicating a required attribute in a request is missing.
#:
#: This is usually a missing attribute in the request data provided for a
#: resource.
MISSING_ATTRIBUTE = WebAPIError(
    code=106,
    msg='Missing value for the attribute',
    error_type='request-attribute-missing',
    http_status=400)


#: An error indicating an extension couldn't be enabled.
ENABLE_EXTENSION_FAILED = WebAPIError(
    code=107,
    msg='There was a problem enabling the extension',
    error_type='extension-enable-error',
    http_status=500)  # 500 Internal Server Error


#: An error indicating an extension couldn't be disabled.
DISABLE_EXTENSION_FAILED = WebAPIError(
    code=108,
    msg='There was a problem disabling the extension',
    error_type='extension-disable-error',
    http_status=500)  # 500 Internal Server Error


#: An error indicating an extension is already installed.
#:
#: This is used when trying to install an extension that is already installed.
EXTENSION_INSTALLED = WebAPIError(
    code=109,
    msg='This extension has already been installed.',
    error_type='extension-already-installed',
    http_status=409)


#: An error indicating an extension installation failed.
INSTALL_EXTENSION_FAILED = WebAPIError(
    code=110,
    msg='An error occurred while installing the extension',
    error_type='extension-install-error',
    http_status=409)


#: An error indicating an item conflicts with another.
#:
#: This is used when attempting to perform a request that would result in a
#: conflict between resources on the server.
DUPLICATE_ITEM = WebAPIError(
    code=111,
    msg='An entry for this item or its unique key(s) already exists',
    error_type='resource-conflict',
    http_status=409)


#: An error indicating an OAuth2 token lacks a required scope.
OAUTH_MISSING_SCOPE_ERROR = WebAPIError(
    code=112,
    msg='Your OAuth2 token lacks the necessary scopes for this request.',
    error_type='auth-oauth2-missing-scope',
    http_status=403,  # 403 Forbidden
)


#: An error indicating an OAuth2 token cannot be used to access the resource.
OAUTH_ACCESS_DENIED_ERROR = WebAPIError(
    code=113,
    msg='OAuth2 token access for this resource is prohibited.',
    error_type='auth-oauth2-access-denied',
    http_status=403,  # 403 Forbidden
)


#: An error indicating the API rate limit has been exceeded.
RATE_LIMIT_EXCEEDED = WebAPIError(
    code=114,
    msg='API rate limit has been exceeded.',
    error_type='request-rate-limit-exceeded',
    http_status=429,  # 429 Too Many Requests
)



"""The base class for an API authentication backend."""

from __future__ import annotations

import logging
import re
from typing import Any, Dict, Optional, Tuple, Union

from django.contrib import auth
from django.http import HttpRequest
from django.utils.translation import gettext as _
from typing_extensions import TypeAlias

from djblets.auth.ratelimit import is_ratelimited
from djblets.webapi.responses import WebAPIResponseHeaders


#: A type alias for authentication results.
#:
#: This was the only return value supported prior to Djblets 3.2.
#:
#: Contents are in the form of:
#:
#: Tuple:
#:     0 (bool):
#:         Whether the authentication request was successful.
#:
#:     1 (str):
#:         The error message to return if authentication failed.
#:
#:         This can be ``None`` if it succeeded, or if it failed
#:         and the default error from
#:         :py:data:`~djblets.webapi.errors.LOGIN_FAILED` should be used.
#:
#:     2 (dict):
#:         Any HTTP headers to return in the response.
#:
#:         This can be ``None`` if no headers need to be returned,
#:         or if it failed and default headers from
#:         :py:data:`~djblets.webapi.errors.LOGIN_FAILED` should be used.
#:
#: Version Added:
#:     3.2
WebAPIAuthenticateResult: TypeAlias = Tuple[bool,
                                            Optional[str],
                                            Optional[WebAPIResponseHeaders]]


#: A type alias for credentials passed to or from auth handlers.
#:
#: Version Added:
#:     3.2
WebAPICredentials: TypeAlias = Dict[str, Any]


#: A type alias for the result of a get_credentials response.
#:
#: Version Added:
#:     3.2
WebAPIGetCredentialsResult: TypeAlias = \
    Optional[Union[WebAPIAuthenticateResult,
                   WebAPICredentials]]


logger = logging.getLogger(__name__)


class WebAPIAuthBackend:
    """Handles a form of authentication for the web API.

    This can be overridden to provide custom forms of authentication, or to
    support multiple types of authentication.

    More than one authentication backend can be used with the web API. In that
    case, the client can make the determination about which to use.

    Auth backends generally need to only override the
    :py:meth:`get_credentials` method, though more specialized ones may
    override other methods as well.

    They must also provide :py:attr:`www_auth_scheme` which is a
    ``WWW-Authenticate`` scheme value.
    """

    #: The auth scheme used in the ``WWW-Authenticate`` header.
    #:
    #: Type:
    #:     str
    www_auth_scheme: Optional[str] = None

    #: A regex of sensitive entries in the credentials dictionary.
    #:
    #: By default, this excludes keys containing "api", "token", "key",
    #: "secret", "password", or "signature" anywhere in the name, in any
    #: casing.
    #:
    #: This can be extended for other sensitive information.
    SENSITIVE_CREDENTIALS_RE = \
        re.compile('api|token|key|secret|password|signature', re.I)

    def get_auth_headers(
        self,
        request: HttpRequest
    ) -> Dict[str, Any]:
        """Return extra authentication headers for the response.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            dict:
            The authentication headers (defaults to empty).
        """
        return {}

    def authenticate(
        self,
        request: HttpRequest,
        **kwargs,
    ) -> Optional[WebAPIAuthenticateResult]:
        """Authenticate a request against this auth backend.

        This will fetch the credentials and attempt an authentication against
        those credentials.

        This function must return ``None`` to indicate it should be skipped
        and another backend should be tried, or a tuple indicating the
        success/failure and additional details for the client.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            tuple or None:
            See :py:class:`WebAPIAuthenticateResult` for details on the format
            for the returned type value.

            If the backend should be skipped, this will return ``None``.
        """
        result: WebAPIAuthenticateResult

        # Check if number of failed login attempts already exceeded
        # before authenticating.
        if is_ratelimited(request, increment=False):
            return False, _('Maximum number of login attempts exceeded.'), None

        credentials = self.get_credentials(request)

        if credentials is None:
            return None

        if isinstance(credentials, dict):
            result = self.login_with_credentials(request, **credentials)
        else:
            assert isinstance(credentials, tuple)
            result = credentials

        if not result[0]:
            # Automatically increment when login with credentials is not
            # successful.
            is_ratelimited(request, increment=True)

        return result

    def get_credentials(
        self,
        request: HttpRequest,
    ) -> WebAPIGetCredentialsResult:
        """Return credentials provided in the request.

        This returns a dictionary of all credentials necessary for this
        auth backend. By default, this expects ``username`` and ``password``,
        though more specialized auth backends may provide other information.
        These credentials will be passed to :py:meth:`login_with_credentials`.

        This function must be implemented by the subclass.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            dict:
            A dictionary of credential information.
        """
        raise NotImplementedError

    def login_with_credentials(
        self,
        request: HttpRequest,
        **credentials,
    ) -> WebAPIAuthenticateResult:
        """Log in against the main authentication backends.

        This takes the provided credentials from the request (as returned by
        :py:meth:`get_credentials`) and attempts a login against the main
        authentication backends used by Django.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            credentials (dict):
                All credentials provided by :py:meth:`get_credentials`.

        Returns:
            tuple or None:
            See the return type in :py:meth:`authenticate`.
        """
        result = self.validate_credentials(request, **credentials)

        if result is not None:
            return result

        log_extra = {
            'request': request,
        }

        cleaned_credentials = self.clean_credentials_for_display(credentials)
        logger.debug(
            'Attempting authentication on API: %s',
            ', '.join([
                '%s=%s' % pair
                for pair in cleaned_credentials.items()
            ]),
            extra=log_extra)

        user = auth.authenticate(request=request, **credentials)

        if user and user.is_active:
            auth.login(request, user)

            return True, None, None

        logger.debug('API Login failed. No valid user found.',
                     extra=log_extra)
        auth.logout(request)

        return False, None, None

    def validate_credentials(
        self,
        request: HttpRequest,
        **credentials,
    ) -> Optional[WebAPIAuthenticateResult]:
        """Validate that credentials are valid.

        This is called before we attempt to authenticate with the credentials,
        and can short-circuit the rest of the authentication process,
        returning a result tuple if desired. If ``None`` is returned,
        authentication proceeds as normal.

        By default, this will attempt to bypass authentication if the
        current user is already logged in and matches the authenticated
        user (if and only if ``username`` appears in the credentials).

        Subclasses can override this to provide more specific behavior for
        their sets of credentials, or to disable this entirely.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            credentials (dict):
                All credentials provided by :py:meth:`get_credentials`.

        Returns:
            tuple or None:
            See the return type in :py:meth:`authenticate`.
        """
        # Don't authenticate if a user is already logged in and the
        # username matches.
        #
        # Note that this does mean that a new password will fail. However,
        # the user is already logged in, and querying the backend for every
        # request is excessive, so it's a tradeoff. The user already has
        # access to the server at this point anyway.
        if (request.user.is_authenticated and
            request.user.get_username() == credentials.get('username')):
            return True, None, None

        return None

    def clean_credentials_for_display(
        self,
        credentials: WebAPICredentials,
    ) -> WebAPICredentials:
        """Clean up a credentials dictionary, removing sensitive information.

        This will take a credentials dictionary and mask anything sensitive,
        preparing it for output to a log file.

        Args:
            credentials (dict):
                A dictionary of credentials provided for authentication.

        Returns:
            dict:
            A sanitized dictionary of credentials, for display purposes.
        """
        clean_credentials = {}

        for key, value in credentials.items():
            if self.SENSITIVE_CREDENTIALS_RE.search(key):
                clean_credentials[key] = '************'
            else:
                clean_credentials[key] = value

        return clean_credentials



"""Base classes for implementing extensions."""

from __future__ import annotations

import locale
import logging
import os
from pathlib import PosixPath
from types import ModuleType
from typing import (Any, Callable, ClassVar, Dict, List, Optional, Sequence,
                    Set, TYPE_CHECKING, Type, Union)

import importlib_resources
from django.conf import settings
from django.contrib.admin.sites import AdminSite
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest
from django.templatetags.static import static
from django.urls import URLPattern, URLResolver, get_mod_func
from django.utils.functional import cached_property
from django.utils.module_loading import import_string
from django.utils.translation import gettext as _
from typing_extensions import NotRequired, TypeAlias, TypedDict

from djblets.extensions.errors import InstallExtensionMediaError
from djblets.extensions.models import RegisteredExtension
from djblets.extensions.settings import ExtensionSettings
from djblets.util.typing import JSONDict
from djblets.webapi.resources import WebAPIResource

if TYPE_CHECKING:
    from importlib_metadata import EntryPoint

    from djblets.extensions.hooks import ExtensionHook
    from djblets.extensions.manager import ExtensionManager


logger = logging.getLogger(__name__)


#: Type for the extension metadata dict.
ExtensionMetadata: TypeAlias = Dict[str, Any]


class BaseStaticBundleConfig(TypedDict):
    """Base class for a bundle configuration.

    This provides the configuration settings allowed by both CSS and
    JavaScript bundles for extensions.

    Version Added:
        5.0
    """

    #: A list of filenames that are included in this bundle.
    #:
    #: It's recommended that this contain a single "index" file that is
    #: responsible for importing other files that comprise the bundle.
    #:
    #: If specifying ``index.ts`` or ``index.js`` the bundle will be treated
    #: as a TypeScript/JavaScript-based bundle, and all imported modules
    #: within the bundle will be resolved and compiled together.
    #:
    #: Type:
    #:     list
    source_filenames: Sequence[str]

    #: A list of URL names that the bundle should be automatically loaded on.
    #:
    #: If provided, the bundle will be selectively loaded only on these pages.
    #:
    #: Type:
    #:     list
    apply_to: NotRequired[Sequence[str]]

    #: Extra options to pass to the compiler.
    #:
    #: This is currently unused by the compilers supported by Djblets.
    #:
    #: Type:
    #:     dict
    compiler_options: NotRequired[dict[str, Any]]

    #: Extra context to include in the generated template.
    #:
    #: For CSS, this will control the attributes of the ``<link>`` tag.
    #: It can be used to offer stylesheets for print or other forms of
    #: media via the ``media`` attribute.
    #:
    #: For JavaScript, this will control the attributes of the ``<script>``
    #: tag. You can, for example, set ``"async"`` or ``"defer"`` to ``True``
    #: to add these attributes.
    #:
    #: Type:
    #:     dict
    extra_context: NotRequired[dict[str, str]]

    #: A list of bundles to include whenever this bundle is included.
    #:
    #: These bundles will be included prior to including this bundle.
    #: They must all be provided by the same extension.
    #:
    #: Type:
    #:     list of str
    include_bundles: NotRequired[Sequence[str]]

    #: The name of the bundle file to generate and load into pages.
    #:
    #: This is expected to be a relative path. The extension support will
    #: take care of placing this file in an appropriate path.
    #:
    #: If not provided, a default filename will be created for the bundle.
    #:
    #: Type:
    #:     str
    output_filename: NotRequired[str]

    #: The name of the Django template used to load this bundle.
    #:
    #: This can be used to provide special logic for loading the bundle.
    #: It's generally not needed.
    template_name: NotRequired[Optional[str]]


class CSSBundleConfig(BaseStaticBundleConfig):
    """CSS static media configuration.

    These provide the options for customizing CSS/LessCSS build behavior.

    See :py:class:`BaseStaticBundleConfig` for more options.

    Version Added:
        5.0
    """

    #: The variant mode used for the CSS.
    #:
    #: This can be set to ``"datauri"`` to encode any referenced images or
    #: fonts inline using ``url("data:...")``.
    variant: NotRequired[Optional[str]]


class JSBundleConfig(BaseStaticBundleConfig):
    """JavaScript static media configuration.

    These provide the options for customizing JavaScript/TypeScript build
    behavior.

    See :py:class:`BaseStaticBundleConfig` for more options.

    Version Added:
        5.0
    """


#: A mapping of CSS bundle names to configurations.
#:
#: Version Added:
#:     5.0
CSSBundleConfigs: TypeAlias = Dict[str, CSSBundleConfig]


#: A mapping of JavaScript bundle names to configurations.
#:
#: Version Added:
#:     5.0
JSBundleConfigs: TypeAlias = Dict[str, JSBundleConfig]


class JSExtension:
    """Base class for a JavaScript extension.

    This can be subclassed to provide the information needed to initialize
    a JavaScript extension.

    The JSExtension subclass is expected to define a :py:attr:`model_class`
    attribute naming its JavaScript counterpart. This would be the variable
    name for the (uninitialized) model for the extension, defined in a
    JavaScript bundle.

    It may also define :py:attr:`apply_to`, which is a list of URL names that
    the extension will be initialized on. If not provided, the extension will
    be initialized on all pages.

    To provide additional data to the model instance, the JSExtension subclass
    can implement :py:meth:`get_model_data` and return a dictionary of data
    to pass. You may also override the :py:meth:`get_settings` method to
    return, a dict of settings to the :py:attr:`model_class`. By default, the
    associated extension's settings are returned.
    """

    #: The name of the JavaScript model class to instantiate.
    #:
    #: This class will be instantiated on the page. It should be a subclass of
    #: :js:class:`Djblets.Extension`.
    model_class: Optional[str] = None

    #: The list of URL names to load this extension on.
    #:
    #: If not provided, this will be loaded on all pages.
    apply_to: Optional[List[str]] = None

    def __init__(
        self,
        extension: Extension,
    ) -> None:
        """Initialize the JavaScript extension.

        Args:
            extension (Extension):
                The main extension that owns this JavaScript extension.
        """
        self.extension = extension

    def applies_to(
        self,
        url_name: str,
    ) -> bool:
        """Return whether this extension applies to the given URL name.

        Args:
            url_name (str):
                The name of the URL.

        Returns:
            bool:
            ``True`` if this JavaScript extension should load on the page
            with the given URL name. ``False`` if it should not load.
        """
        return self.apply_to is None or url_name in self.apply_to

    def get_model_data(
        self,
        request: HttpRequest,
        **kwargs,
    ) -> JSONDict:
        """Return model data for the Extension model instance in JavaScript.

        Subclasses can override this to return custom data to pass to
        the extension class defined in :js:attr:`model_class`. This data must
        be JSON-serializable.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            dict:
            Model data to pass to the constructor of the JavaScript extension
            class.
        """
        return {}

    def get_settings(self) -> JSONDict:
        """Return the settings for the JS Extension.

        By default, this is the associated :py:class:`Extension` object's
        settings. Subclasses may override this method to provide different
        settings.

        These settings will be provided to the :py:attr:`model_class` as a
        ``settings`` key in its initialization options.

        Returns:
            dict:
            The extension settings.
        """
        return self.extension.settings


class Extension:
    """Base class for an extension.

    Extensions must subclass this class. They'll automatically have support for
    settings, adding hooks, and plugging into the administration UI.

    For information on writing extensions, see :ref:`writing-extensions`.
    """

    #: The ID of the extension.
    id: ClassVar[str]

    #: Information and metadata about the extension.
    info: ClassVar[ExtensionInfo]

    #: The instance of this extension created by the extension manager.
    #:
    #: Type:
    #:     Extension
    instance: Optional[Extension] = None

    #: The extension registration in the database.
    #:
    #: Type:
    #:     djblets.extensions.models.RegisteredExtension
    registration: RegisteredExtension

    #: Metadata describing the package.
    #:
    #: This is used to set the name, version, and other information for the
    #: extension. This data defaults to coming from the extension's Python
    #: package metadata, but can be set on the extension itself. You would
    #: want to use this if shipping multiple extensions in a single package,
    #: if you want to include a space in the package name, or if you want
    #: the version to be set independently of the package, for example.
    #:
    #: The following keys are supported:
    #:
    #: * ``Name``: The displayed name of the extension.
    #: * ``Version``: The version of the extension.
    #: * ``Summary``: A summary of the extension.
    #: * ``Description``: A more detailed description of the extension.
    #: * ``License``: The license the extension was released under.
    #: * ``Author``: The name of the author writing/maintaining the extension.
    #: * ``Author-email``: The e-mail address of the author writing/maintaining
    #:   the extension.
    #: * ``Author-home-page``: The URL of the author writing/maintaining the
    #:   extension.
    #: * ``Home-page``: The URL of the extension's home/product/company page.
    metadata: Optional[ExtensionMetadata] = None

    #: Whether or not the extension is user-configurable.
    #:
    #: If ``True``, the extension will have a :guilabel:`Configure` link in
    #: the extension list that will take the user to a page for modifying
    #: extension settings. The extension must provide a :file:`admin_urls.py`
    #: file defining a top-level URL (:regexp:`^$`) pointing to a view that
    #: will handle configuration.
    is_configurable: bool = False

    #: Default settings for the extension.
    #:
    #: These values will be used when looking up keys in :py:attr:`settings`
    #: that don't have custom values saved in the database.
    default_settings: JSONDict = {}

    #: Whether or not the extension has a database administration site.
    #:
    #: If ``True``, the extension will have a :guilabel:`Database` link in
    #: the extension list that will take the user to a page for adding/editing
    #: any database models registered by the extension.
    has_admin_site: bool = False

    #: A list of any extension IDs to enable when this extension is enabled.
    #:
    #: This is used to ensure that another extension is enabled before this
    #: one. It's primarily for extensions that are augmenting or depending on
    #: another extension's functionality.
    requirements: List[str] = []

    #: A list of API resource instances offered by this extension.
    #:
    #: Each entry in the list is an instance of a custom
    #: :py:class:`~djblets.webapi.resources.WebAPIResource`. These resources
    #: will appear underneath the extension's own resource.
    resources: List[WebAPIResource] = []

    #: A list of Django application module paths to load.
    #:
    #: Each of these will be added to :django:setting:`INSTALLED_APPS` while
    #: the extension is enabled. It follows the same format as that setting.
    apps: List[str] = []

    #: A list of Django context processors to load.
    #:
    #: Each of these will be added to
    #: :django:setting:`TEMPLATE_CONTEXT_PROCESSORS` while the extension is
    #: enabled. It follows the same format as that setting.
    context_processors: List[str] = []

    #: A list of Django middleware to load.
    #:
    #: Each of these will be run as if they were part of
    #: :django:setting:`MIDDLEWARE` depending on your setup) while the
    #: extension is enabled. It follows the same format as that setting.
    middleware: List[str] = []

    #: A dictionary of CSS bundles to include in the package.
    #:
    #: These will be automatically packaged along with the extension, and can
    #: be loaded in templates using :py:func:`{% ext_css_bundle %}
    #: <djblets.extensions.templatetags.djblets_extensions.ext_css_bundle>`.
    #:
    #: Each key in the dictionary is the name of the bundle, and the
    #: value is a dictionary containing a ``source_filenames`` key pointing
    #: to a list of CSS/LessCSS files.
    #:
    #: A special bundle ID of ``default`` will cause the CSS to be loaded on
    #: every page.
    #:
    #: Version Changed:
    #:     5.0:
    #:     Added specific typing for bundle configuration. Previous versions
    #:     allowed arbitrary dictionaries to be used.
    css_bundles: CSSBundleConfigs = {}

    #: A dictionary of JavaScript bundles to include in the package.
    #:
    #: These will be automatically packaged along with the extension, and can
    #: be loaded in templates using :py:func:`{% ext_js_bundle %}
    #: <djblets.extensions.templatetags.djblets_extensions.ext_js_bundle>`.
    #:
    #: Each key in the dictionary is the name of the bundle, and the
    #: value is a dictionary containing a ``source_filenames`` key pointing
    #: to a list of JavaScript files.
    #:
    #: A special bundle ID of ``default`` will cause the JavaScript to be
    #: loaded on every page.
    #:
    #: Version Changed:
    #:     5.0:
    #:     Added specific typing for bundle configuration. Previous versions
    #:     allowed arbitrary dictionaries to be used.
    js_bundles: JSBundleConfigs = {}

    #: A list of JavaScript extension classes to enable on pages.
    #:
    #: Each entry in the list is a :py:class:`JSExtension` subclass to load.
    js_extensions: Sequence[type[JSExtension]] = []

    ######################
    # Instance variables #
    ######################

    #: The hooks currently registered and enabled for the extension.
    #:
    #: Type:
    #:     set of djblets.extensions.hooks.ExtensionHook
    hooks: Set[ExtensionHook]

    #: URLs for the extension's admin interface.
    #:
    #: Type:
    #:     list of django.urls.URLResolver
    admin_urlpatterns: List[URLResolver]

    #: The database administration site set for the extension.
    #:
    #: This will be set automatically if :py:attr:`has_admin_site`` is
    #: ``True``.
    #:
    #: Type:
    #:     django.contrib.admin.sites.AdminSite
    admin_site: Optional[AdminSite]

    #: URLs for the extension's admin site (DB forms, etc).
    #:
    #: Type:
    #:     list of django.urls.URLPattern
    admin_site_urlpatterns: List[Union[URLPattern, URLResolver]]

    #: The extension manager that manages this extension.
    #:
    #: Type:
    #:     djblets.extensions.manager.ExtensionManager
    extension_manager: ExtensionManager

    #: The list of middleware classses.
    #:
    #: Version Added:
    #:     2.2.4
    #:
    #: Type:
    #:     list of callable
    middleware_classes: Sequence[Callable]

    #: The settings for the extension.
    #:
    #: Type:
    #:     djblets.extensions.settings.ExtensionSettings
    settings: ExtensionSettings

    def __init__(
        self,
        extension_manager: ExtensionManager,
    ) -> None:
        """Initialize the extension.

        Subclasses should not override this. Instead, they should override
        :py:meth:`initialize`.

        Args:
            extension_manager (djblets.extensions.manager.ExtensionManager):
                The extension manager that manages this extension.
        """
        self.extension_manager = extension_manager
        self.hooks = set()
        self.settings = ExtensionSettings(self)
        self.admin_site = None

        self.middleware_classes = []

        for middleware_path in self.middleware:
            self.middleware_classes.append(import_string(middleware_path))

        self.initialize()

    def initialize(self) -> None:
        """Initialize the extension.

        Subclasses can override this to provide any custom initialization.
        They do not need to call the parent function, as it does nothing.
        """
        pass

    def shutdown(self) -> None:
        """Shut down the extension.

        By default, this calls shutdown_hooks.

        Subclasses should override this if they need custom shutdown behavior.
        """
        self.shutdown_hooks()

    def shutdown_hooks(self) -> None:
        """Shut down all hooks for the extension."""
        for hook in self.hooks.copy():
            if hook.initialized:
                hook.disable_hook()

    def get_static_url(
        self,
        path: str,
    ) -> str:
        """Return the URL to a static media file for this extension.

        This takes care of resolving the static media path name to a path
        relative to the web server. If a versioned media file is found, it
        will be used, so that browser-side caching can be used.

        Args:
            path (str):
                The path within the static directory for the extension.

        Returns:
            str:
            The resulting static media URL.
        """
        return static('ext/%s/%s' % (self.id, path))

    def get_bundle_id(
        self,
        name: str,
    ) -> str:
        """Return the ID for a CSS or JavaScript bundle.

        This ID should be used when manually referencing the bundle in a
        template. The ID will be unique across all extensions.

        Args:
            name (str):
                The name of the bundle to reference.

        Returns:
            str:
            The ID of the bundle corresponding to the name.
        """
        return '%s-%s' % (self.id, name)

    @cached_property
    def admin_urlconf(self) -> ModuleType:
        """The module defining URLs for the extension's admin site."""
        name = '%s.%s' % (get_mod_func(self.__class__.__module__)[0],
                          'admin_urls')

        try:
            return __import__(name, {}, {}, [''])
        except Exception as e:
            raise ImproperlyConfigured(
                "Error while importing extension's admin URLconf %r: %s" %
                (name, e))


class ExtensionInfo:
    """Information on an extension.

    This class stores the information and metadata on an extension. This
    includes the name, version, author information, where it can be downloaded,
    whether or not it's enabled or installed, and anything else that may be
    in the Python package for the extension.
    """

    #: Encodings to try using to decode metadata.
    #:
    #: Deprecated:
    #:     3.3:
    #:     This is no longer used by Djblets and may be removed in a future
    #:     version.
    encodings: List[str] = [
        'utf-8',
        locale.getpreferredencoding(False),
        'latin1',
    ]

    ######################
    # Instance variables #
    ######################

    #: A list of any extension IDs to enable when this extension is enabled.
    #:
    #: This is used to ensure that another extension is enabled before this
    #: one. It's primarily for extensions that are augmenting or depending on
    #: another extension's functionality.
    #:
    #: Type:
    #:     list of Extension
    requirements: List[Type[Extension]]

    @classmethod
    def create_from_entrypoint(
        cls,
        entrypoint: EntryPoint,
        ext_class: Type[Extension],
    ) -> ExtensionInfo:
        """Create a new ExtensionInfo from a Python EntryPoint.

        This will pull out information from the EntryPoint and return a new
        ExtensionInfo from it.

        It handles pulling out metadata from the older :file:`PKG-INFO` files
        and the newer :file:`METADATA` files.

        Args:
            entrypoint (importlib.metadata.EntryPoint):
                The EntryPoint pointing to the extension class.

            ext_class (type):
                The extension class (subclass of :py:class:`Extension`).

        Returns:
            ExtensionInfo:
            An ExtensionInfo instance, populated with metadata from the
            package.
        """
        assert entrypoint.dist

        # This is a 'PackageMetadata' protocol, which gives us some basic
        # primitives for how to safely iterate over contents. This is not a
        # simple dictionary under the hood.
        dist_metadata = entrypoint.dist.metadata
        metadata = {
            key: dist_metadata[key]
            for key in dist_metadata
        }

        return cls(ext_class=ext_class,
                   package_name=metadata.get('Name', ext_class.id),
                   metadata=metadata)

    def __init__(
        self,
        ext_class: Type[Extension],
        package_name: str,
        metadata: ExtensionMetadata = {},
    ) -> None:
        """Instantiate the ExtensionInfo using metadata and an extension class.

        This will set information about the extension based on the metadata
        provided by the caller and the extension class itself.

        Args:
            ext_class (type):
                The extension class (subclass of :py:class:`Extension`).

            package_name (str):
                The package name owning the extension.

            metadata (ExtensionMetadata, optional):
                Optional metadata for the extension. If the extension provides
                its own metadata, that will take precedence.

        Raises:
            TypeError:
                The parameters passed were invalid (they weren't a new-style
                call or a legacy entrypoint-related call).
        """
        try:
            issubclass(ext_class, Extension)
        except TypeError:
            logger.error('Unexpected parameters passed to '
                         'ExtensionInfo.__init__: ext_class=%r, '
                         'package_name=%r, metadata=%r',
                         ext_class, package_name, metadata)

            raise TypeError(
                _('Invalid parameters passed to ExtensionInfo.__init__'))

        # Set the base information from the extension and the package.
        self.package_name = package_name
        self.module_name = ext_class.__module__
        self.app_name = '.'.join(ext_class.__module__.split('.')[:-1])
        self.is_configurable = ext_class.is_configurable
        self.has_admin_site = ext_class.has_admin_site
        self.installed_htdocs_path = \
            os.path.join(settings.MEDIA_ROOT, 'ext', self.package_name)
        self.installed_static_path = \
            os.path.join(settings.STATIC_ROOT, 'ext', ext_class.id)

        # State set by ExtensionManager.
        self.enabled = False
        self.installed = False
        self.requirements = []
        self.apps_registered = False
        self.context_processors_registered = False

        # Set information from the provided metadata.
        if ext_class.metadata is not None:
            metadata.update(ext_class.metadata)

        self.metadata = metadata
        self.name = metadata.get('Name', package_name)
        self.version = metadata.get('Version')
        self.summary = metadata.get('Summary')
        self.description = metadata.get('Description')
        self.author = metadata.get('Author')
        self.author_email = metadata.get('Author-email')
        self.license = metadata.get('License')
        self.url = metadata.get('Home-page')
        self.author_url = metadata.get('Author-home-page', self.url)

    @cached_property
    def installed_static_version_path(self) -> str:
        """The path to the static media version file.

        This file records the version of the extension used when last
        installing the static media files.

        Type:
            str
        """
        return os.path.join(self.installed_static_path, '.version')

    def has_resource(
        self,
        path: str,
    ) -> bool:
        """Return whether an extension has a resource in its package.

        A resource is a file or directory that exists within an extension's
        package.

        Args:
            path (str):
                The ``/``-delimited path to the resource within the package.

        Returns:
            bool:
            ``True`` if the resource exits. ``False`` if it does not.
        """
        return (
            importlib_resources.files(self.module_name) /
            PosixPath(path)
        ).exists()

    def extract_resource(
        self,
        path: str,
    ) -> Optional[str]:
        """Return the filesystem path to an extracted resource.

        A resource is a file or directory that exists within an extension's
        package.

        This will extract the resource from the package, if the package is
        compressed, and then return the local path to the file on the
        filesystem.

        Args:
            path (str):
                The ``/``-delimited path to the resource within the package.

        Returns:
            str:
            The local filesystem path to the resource, or ``None`` if it
            could not be found.
        """
        if self.has_resource(path):
            # Note that we're only supporting static files that exist on the
            # filesystem, and aren't in some hypothetical zipped package or
            # something. This is safe for Wheels.
            #
            # If we didn't want to make this assumption, we'd need to use
            # importlib.resources.as_file(), which would set up a temp
            # directory in either case. We don't want that, so we'll keep
            # our assumption.
            return str(importlib_resources.files(self.module_name) /
                       PosixPath(path))

        return None

    def write_installed_static_version(self) -> None:
        """Write the extension's current static media version to disk.

        This will write the extension's current version in its static media
        directory, creating that directory if necessary. This will allow
        the extension manager to check if new media files need to be installed.

        Raises:
            djblets.extensions.errors.InstallExtensionMediaError:
                There was an error writing the version to the static media
                directory. Details are in the error message.
        """
        version_path = self.installed_static_version_path
        parent_path = os.path.dirname(version_path)

        try:
            if not os.path.exists(parent_path):
                os.makedirs(parent_path, 0o755)

            with open(version_path, 'w') as fp:
                fp.write('%s\n' % self.version)
        except Exception:
            raise InstallExtensionMediaError(
                _('Unable to write the extension static media version '
                  'file "%(path)s". Please make sure the file and its parent '
                  'directory are owned by the web server.')
                % {
                    'path': version_path,
                })

    def get_installed_static_version(self) -> Optional[str]:
        """Return the extension's locally-written static media version.

        Returns:
            str:
            The extension version written to disk, or ``None`` if it didn't
            exist or couldn't be read.
        """
        try:
            with open(self.installed_static_version_path, 'r') as fp:
                return fp.read().strip()
        except IOError:
            return None

    def __str__(self) -> str:
        return '%s %s (enabled = %s)' % (self.name, self.version, self.enabled)



"""Specialized HTTP response classes.

Version Added:
    4.0
"""

from __future__ import annotations

from typing import Callable, Iterator, List, Optional, TYPE_CHECKING, Union

from django.http.response import StreamingHttpResponse
from typing_extensions import NotRequired, TypeAlias, TypedDict

if TYPE_CHECKING:
    from django.http import HttpRequest


class EventStreamMessage(TypedDict):
    """A message in an event stream.

    One message may contain an ID, application-defined event type, serialized
    payload data, and a retry time in milliseconds. These are all optional,
    but any empty message will be skipped.

    Version Added:
        4.0
    """

    #: The payload data to send.
    #:
    #: This will be sent as the ``data`` field.
    #:
    #: Type:
    #:     bytes or str
    data: NotRequired[Optional[Union[bytes, str]]]

    #: The event name to send.
    #:
    #: This will be sent as the ``event`` field.
    #:
    #: Type:
    #:     str
    event: NotRequired[Optional[str]]

    #: The ID to send.
    #:
    #: This will be sent as the ``id`` field.
    #:
    #: Type:
    #:     str
    id: NotRequired[Optional[str]]

    #: The retry time in milliseconds for reconnections after a disconnect.
    #:
    #: This will be sent as the ``retry`` field.
    #:
    #: Type:
    #:     int
    retry_ms: NotRequired[Optional[int]]


#: An iterator for event stream messages.
#:
#: Version Added:
#:     4.0
EventStreamMessages: TypeAlias = Iterator[EventStreamMessage]


#: An event stream that generates messages.
#:
#: This may be an iterator of messages, or a callable that takes an optional
#: :mailheader:`Last-Event-ID` header value and then yields messages.
#:
#: Version Added:
#:     4.0
EventStream: TypeAlias = Union[
    EventStreamMessages,
    Callable[[Optional[str]], EventStreamMessages],
]


class EventStreamHttpResponse(StreamingHttpResponse):
    """A Server-Sent Events (SSE) HTTP response.

    This manages a response containing Server-Sent Events, which can be
    used to continuously stream content (events, data, IDs, and retry
    intervals) to a caller in a standardized way.

    Server-Sent Events are useful for long-running, resumeable process, such
    as conveying progress for long-running but interruptible operations,
    new updates to content, or events on a message bus.

    Providers of SSE responses can use the :mailheader:`Last-Event-ID` header
    value to determine where the client left off, and resume from that point.
    If a function is passed as the event stream source, it will receive any
    value in this header as an argument.

    Note:
        On HTTP/1.1 connections, web browsers are limited to 6 concurrent HTTP
        connections to the same domain across all tabs, and a long-running SSE
        stream will consume one of those slots until it ends.

        HTTP/2 connections have a higher limit.

        For widest compatibility when serving over HTTP/1.1 connections, it's
        recommended to use shorter-lived, resumeable SSE responses.

    Version Added:
        4.0
    """

    streaming = True

    ######################
    # Instance variables #
    ######################

    #: The event stream providing values for the response.
    #:
    #: Type:
    #:     callable or generator
    event_stream: EventStream

    #: The HTTP request from the client, if provided to the response.
    #:
    #: Type:
    #:     django.http.HttpRequest
    request: Optional[HttpRequest]

    def __init__(
        self,
        event_stream: EventStream,
        *,
        request: Optional[HttpRequest] = None,
        status: int = 200,
        content_type: str = 'text/event-stream',
        **kwargs,
    ) -> None:
        """Initialize the event stream.

        The provided event stream source will be processed when sending
        content to the client, and not before.

        Args:
            event_stream (callable or generator):
                The event stream as a generator or a callable yielding
                results.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

            status (int, optional):
                The HTTP status code to send in the response.

            content_type (str, optional):
                The content type to send in the response.

                For greatest compatibility, this should be left as the
                default of :mimetype:`text/event-stream`.

            **kwargs (dict):
                Additional keyword arguments to pass to the parent
                constructor.
        """
        self.event_stream = event_stream
        self.request = request

        super().__init__(self._process_events(),
                         status=status,
                         content_type=content_type,
                         **kwargs)

        self['Cache-Control'] = 'no-cache'

        # It's important that we disable any encodings, or the GZip middleware
        # will try to compress multiple messages at once and delay sending
        # them until there's a suitably-compressed segment. This happens in
        # GZipMiddleware and in django.utils.text.compress_sequence.
        #
        # Ideally we'd just tell it to flush per-message, but we can't. We
        # can do our own compression if the middleware is installed, manually
        # handling all compression via django.utils.text.compress_string()
        # if `Accept-Encoding:` includes "gzip", and that's worth considering
        # in the future, but may not ultimately be worth it given the size
        # of the small messages being sent.
        self['Content-Encoding'] = ''

    def _process_events(self) -> Iterator[bytes]:
        """Generate serialized messages based on the event stream.

        This will iterate through the response's event stream, converting
        any messages to serialized Server-Sent Event data.

        This can only be called once per response instance.

        Yields:
            bytes:
            Each serialized message to send to the client.
        """
        make_bytes = self.make_bytes
        event_stream = self.event_stream

        if callable(event_stream):
            last_id: Optional[str]
            request = self.request

            if request is not None:
                last_id = request.META.get('HTTP_LAST_EVENT_ID')
            else:
                last_id = None

            event_stream = event_stream(last_id)

        for message in event_stream:
            if message:
                data = message.get('data')
                event = message.get('event')
                message_id = message.get('id')
                retry = message.get('retry_ms')

                payload: List[bytes] = []

                if message_id is not None:
                    if message_id == '':
                        # This will reset the ID for the Last-Event-ID header.
                        payload.append(b'id\n')
                    else:
                        payload.append(b'id: %s\n' % make_bytes(message_id))

                if event is not None:
                    payload.append(b'event: %s\n' % make_bytes(event))

                if data is not None:
                    payload += [
                        b'data: %s\n' % make_bytes(line)
                        for line in data.splitlines()
                    ]

                if retry is not None:
                    payload.append(b'retry: %d\n' % retry)

                payload.append(b'\n')

                yield b''.join(payload)
            else:
                # This is a no-op, but will keep the connection open.
                yield b':\n\n'



"""Template loaders for extensions."""

from __future__ import annotations

from pathlib import PosixPath

import importlib_resources
from django.template import Origin, TemplateDoesNotExist
from django.template.loaders.base import Loader as BaseLoader

from djblets.extensions.manager import get_extension_managers


class ExtensionOrigin(Origin):
    """An origin for a template in an extension.

    Version Added:
        3.0
    """

    def __init__(self, package, resource, *args, **kwargs):
        """Initialize the origin.

        Args:
            package (unicode):
                The name of the package providing the template.

            resource (unicode):
                The resource path within the package.

            *args (tuple):
                Positional arguments to pass to the parent.

            **kwargs (dict):
                Keyword arguments to pass to the parent.
        """
        self.package = package
        self.resource = resource

        super(ExtensionOrigin, self).__init__(*args, **kwargs)


class Loader(BaseLoader):
    """Loads templates found within an extension.

    This will look through all enabled extensions and attempt to fetch
    the named template under the :file:`templates` directory within the
    extension's package.

    This should be added last to the list of template loaders.

    .. versionadded:: 0.9
    """

    is_usable = True

    def get_contents(
        self,
        origin: ExtensionOrigin,
    ) -> str:
        """Return the contents of a template.

        Args:
            origin (ExtensionOrigin):
                The origin of the template.

        Returns:
            str:
            The resulting template contents.

        Raises:
            TemplateDoesNotExist:
                The template could not be found.
        """
        try:
            path = (importlib_resources.files(origin.package) /
                    PosixPath(origin.resource))

            return path.read_text()
        except Exception:
            raise TemplateDoesNotExist(origin)

    def get_template_sources(self, template_name):
        """Load templates from enabled extensions.

        Args:
            template_name (unicode):
                The name of the template to load.

        Yields:
            ExtensionOrigin:
            Each possible location for the template.
        """
        resource = f'templates/{template_name}'

        for extmgr in get_extension_managers():
            for ext in extmgr.get_enabled_extensions():
                package = ext.info.app_name

                yield ExtensionOrigin(
                    package=package,
                    resource=resource,
                    name=f'extension:{package}:{resource}',
                    template_name=template_name,
                    loader=self)



"""A pipeline compiler for TypeScript.

Version Added:
    4.0
"""

from pipeline.compilers.es6 import ES6Compiler


class TypeScriptCompiler(ES6Compiler):
    """A pipeline compiler which compiles .ts to .js.

    To use this, add the path to this class to
    ``settings.PIPELINE['COMPILERS']``.

    Version Added:
        4.0
    """

    input_extension = 'ts'

    def match_file(
        self,
        path: str,
    ) -> bool:
        """Return whether the given path should use this compiler.

        Args:
            path (str):
                The source path.

        Returns:
            bool:
            Whether this compiler should be used for the given path.
        """
        return path.endswith(self.input_extension)



"""Unit tests for djblets.db.query_comparator.

Version Added:
    3.4
"""

from __future__ import annotations

import re
from typing import TYPE_CHECKING

from django.contrib.auth.models import Group, User
from django.db.models import Count, Exists, F, OuterRef, Q, Value

from djblets.db.query_catcher import CatchQueriesContext
from djblets.db.query_comparator import (compare_queries,
                                         _format_node,
                                         _normalize_q)
from djblets.testing.testcases import TestCase, TestModelsLoaderMixin

if TYPE_CHECKING:
    from djblets.db.query_comparator import ExpectedQueries


class CompareQueriesTests(TestModelsLoaderMixin, TestCase):
    """Unit tests for djblets.db.query_comparator.compare_queries.

    Version Added:
        3.4
    """

    maxDiff = None

    _extra_ws_re = re.compile(r'\s{2,}')

    def test_with_select_q_and_match(self) -> None:
        """Testing compare_queries with SELECT and Q() and match"""
        queries: ExpectedQueries = [
            {
                'model': User,
                'where': Q(is_active=True),
            },
        ]

        with compare_queries(queries) as ctx:
            list(User.objects.filter(is_active=True))

        self.assertFalse(ctx['has_mismatches'])
        self.assertEqual(ctx['num_executed_queries'], 1)
        self.assertEqual(ctx['num_expected_queries'], 1)
        self.assertFalse(ctx['query_count_mismatch'])
        self.assertEqual(ctx['query_mismatches'], [])

    def test_with_match_complex(self) -> None:
        """Testing compare_queries with match and complex query"""
        queries: ExpectedQueries = [
            {
                'annotations': {'__count': Count('*')},
                'model': User,

                'subqueries': [
                    {
                        'distinct': True,
                        'extra': {
                            'username': (
                                'select username from auth_user where id=%s',
                                [0],
                            ),
                        },
                        'extra_tables': ('auth_group',),
                        'join_types': {
                            'djblets_privacy_storedconsentdata': 'INNER JOIN',
                        },
                        'model': User,
                        'num_joins': 1,
                        'only_fields': {'username'},
                        'subquery': True,
                        'tables': {
                            'auth_user',
                            'djblets_privacy_storedconsentdata',
                        },
                        'where': Q(~Q(__Exists__subquery__=1) &
                                   Q(storedconsentdata__pk__gt=0)),

                        'subqueries': [
                            {
                                'annotations': {
                                    'a': Value(1),
                                },
                                'limit': 2,
                                'model': User,
                                'offset': 1,
                                'subquery': True,
                                'values_select': ('pk',),
                                'where': Q(Q(pk=OuterRef('pk')) &
                                           Q(username='test')),
                            },
                        ],
                    },
                ],
            },
        ]

        with compare_queries(queries) as ctx:
            count = (
                User.objects
                .filter(
                    ~Exists(
                        User.objects
                        .filter(Q(pk=OuterRef('pk')) &
                                Q(username='test'))
                        .values('pk')[1:2]
                    ) &
                    Q(storedconsentdata__pk__gt=0)
                )
                .select_related('storedconsentdata')
                .extra(
                    select={
                        'username': ('select username from auth_user'
                                     ' where id=%s'),
                    },
                    select_params=(0,),
                    tables=['auth_group'])
                .order_by('username')
                .only('username')
                .distinct()
                .count()
            )

        self.assertEqual(count, 0)
        self.assertEqual(
            ctx,
            {
                '_unchecked_mismatched_attrs': set(),
                'has_mismatches': False,
                'num_executed_queries': 1,
                'num_expected_queries': 1,
                'query_count_mismatch': False,
                'query_mismatches': [],
            })

    def test_with_mismatch_distinct(self) -> None:
        """Testing compare_queries with mismatch and distinct"""
        queries: ExpectedQueries = [
            {
                'model': User,
            },
        ]

        with compare_queries(queries) as ctx:
            list(User.objects.distinct())

        self.assertTrue(ctx['has_mismatches'])
        self.assertEqual(ctx['num_executed_queries'], 1)
        self.assertEqual(ctx['num_expected_queries'], 1)
        self.assertFalse(ctx['query_count_mismatch'], 1)

        mismatches = ctx['query_mismatches']
        self.assertEqual(len(mismatches), 1)

        mismatch = mismatches[0]
        self.assertIsNotNone(mismatch['executed_query'])
        self.assertEqual(mismatch['index'], 0)
        self.assertIsNone(mismatch['note'])
        self.assertIsNotNone(mismatch['traceback'])
        self.assertEqual(
            mismatch['subqueries'],
            {
                '_unchecked_mismatched_attrs': set(),
                'has_mismatches': False,
                'num_executed_queries': 0,
                'num_expected_queries': 0,
                'query_count_mismatch': False,
                'query_mismatches': [],
            })
        self.assertEqual(
            mismatch['mismatched_attrs'],
            [
                {
                    'executed_value': 'True',
                    'expected_value': 'False',
                    'name': 'distinct',
                    'raw_executed_value': True,
                    'raw_expected_value': False,
                },
            ])

    def test_with_mismatch_extra(self) -> None:
        """Testing compare_queries with mismatch and extra"""
        queries: ExpectedQueries = [
            {
                'model': User,
            },
        ]

        with compare_queries(queries) as ctx:
            list(User.objects.extra(
                select={
                    'username': 'select username from auth_user where id=%s',
                },
                select_params=(0,)))

        self.assertTrue(ctx['has_mismatches'])
        self.assertEqual(ctx['num_executed_queries'], 1)
        self.assertEqual(ctx['num_expected_queries'], 1)
        self.assertFalse(ctx['query_count_mismatch'], 1)

        mismatches = ctx['query_mismatches']
        self.assertEqual(len(mismatches), 1)

        mismatch = mismatches[0]
        self.assertIsNotNone(mismatch['executed_query'])
        self.assertEqual(mismatch['index'], 0)
        self.assertIsNone(mismatch['note'])
        self.assertIsNotNone(mismatch['traceback'])
        self.assertEqual(
            mismatch['subqueries'],
            {
                '_unchecked_mismatched_attrs': set(),
                'has_mismatches': False,
                'num_executed_queries': 0,
                'num_expected_queries': 0,
                'query_count_mismatch': False,
                'query_mismatches': [],
            })
        self.assertEqual(
            mismatch['mismatched_attrs'],
            [
                {
                    'executed_value': (
                        "{'username': ('select username from auth_user"
                        " where id=%s', [0])}"
                    ),
                    'expected_value': '{}',
                    'name': 'extra',
                    'raw_executed_value': {
                        'username': (
                            'select username from auth_user where id=%s',
                            [0],
                        ),
                    },
                    'raw_expected_value': {},
                },
            ])

    def test_with_mismatch_extra_order_by(self) -> None:
        """Testing compare_queries with mismatch and extra_order_by"""
        queries: ExpectedQueries = [
            {
                'model': User,
            },
        ]

        with compare_queries(queries) as ctx:
            list(User.objects.extra(order_by=['-pk']))

        self.assertTrue(ctx['has_mismatches'])
        self.assertEqual(ctx['num_executed_queries'], 1)
        self.assertEqual(ctx['num_expected_queries'], 1)
        self.assertFalse(ctx['query_count_mismatch'], 1)

        mismatches = ctx['query_mismatches']
        self.assertEqual(len(mismatches), 1)

        mismatch = mismatches[0]
        self.assertIsNotNone(mismatch['executed_query'])
        self.assertEqual(mismatch['index'], 0)
        self.assertIsNone(mismatch['note'])
        self.assertIsNotNone(mismatch['traceback'])
        self.assertEqual(
            mismatch['subqueries'],
            {
                '_unchecked_mismatched_attrs': set(),
                'has_mismatches': False,
                'num_executed_queries': 0,
                'num_expected_queries': 0,
                'query_count_mismatch': False,
                'query_mismatches': [],
            })
        self.assertEqual(
            mismatch['mismatched_attrs'],
            [
                {
                    'executed_value': "['-pk']",
                    'expected_value': '()',
                    'name': 'extra_order_by',
                    'raw_executed_value': ['-pk'],
                    'raw_expected_value': (),
                },
            ])

    def test_with_mismatch_extra_tables(self) -> None:
        """Testing compare_queries with mismatch and extra_tables"""
        queries: ExpectedQueries = [
            {
                'model': User,
            },
        ]

        with compare_queries(queries) as ctx:
            list(User.objects.extra(tables=['auth_group']))

        self.assertTrue(ctx['has_mismatches'])
        self.assertEqual(ctx['num_executed_queries'], 1)
        self.assertEqual(ctx['num_expected_queries'], 1)
        self.assertFalse(ctx['query_count_mismatch'], 1)

        mismatches = ctx['query_mismatches']
        self.assertEqual(len(mismatches), 1)

        mismatch = mismatches[0]
        self.assertIsNotNone(mismatch['executed_query'])
        self.assertEqual(mismatch['index'], 0)
        self.assertIsNone(mismatch['note'])
        self.assertIsNotNone(mismatch['traceback'])
        self.assertEqual(
            mismatch['subqueries'],
            {
                '_unchecked_mismatched_attrs': set(),
                'has_mismatches': False,
                'num_executed_queries': 0,
                'num_expected_queries': 0,
                'query_count_mismatch': False,
                'query_mismatches': [],
            })
        self.assertEqual(
            mismatch['mismatched_attrs'],
            [
                {
                    'executed_value': "('auth_group',)",
                    'expected_value': '()',
                    'name': 'extra_tables',
                    'raw_executed_value': ('auth_group',),
                    'raw_expected_value': (),
                },
            ])

    def test_with_mismatch_joins(self) -> None:
        """Testing compare_queries with mismatch and joins"""
        queries: ExpectedQueries = [
            {
                'model': User,
            },
        ]

        with compare_queries(queries) as ctx:
            list(
                User.objects
                .filter(Q(storedconsentdata__pk__gt=0))
            )

        self.assertTrue(ctx['has_mismatches'])
        self.assertEqual(ctx['num_executed_queries'], 1)
        self.assertEqual(ctx['num_expected_queries'], 1)
        self.assertFalse(ctx['query_count_mismatch'], 1)

        mismatches = ctx['query_mismatches']
        self.assertEqual(len(mismatches), 1)

        mismatch = mismatches[0]
        self.assertIsNotNone(mismatch['executed_query'])
        self.assertEqual(mismatch['index'], 0)
        self.assertIsNone(mismatch['note'])
        self.assertIsNotNone(mismatch['traceback'])
        self.assertEqual(
            mismatch['subqueries'],
            {
                '_unchecked_mismatched_attrs': set(),
                'has_mismatches': False,
                'num_executed_queries': 0,
                'num_expected_queries': 0,
                'query_count_mismatch': False,
                'query_mismatches': [],
            })
        self.assertEqual(
            mismatch['mismatched_attrs'],
            [
                {
                    'executed_value': '1',
                    'expected_value': '0',
                    'name': 'num_joins',
                    'raw_executed_value': 1,
                    'raw_expected_value': 0,
                },
                {
                    'executed_value': (
                        "{'auth_user', 'djblets_privacy_storedconsentdata'}"
                    ),
                    'expected_value': "{'auth_user'}",
                    'name': 'tables',
                    'raw_executed_value': {
                        'auth_user',
                        'djblets_privacy_storedconsentdata',
                    },
                    'raw_expected_value': {'auth_user'},
                },
                {
                    'executed_value': (
                        "{'djblets_privacy_storedconsentdata':"
                        " 'INNER JOIN'}"
                    ),
                    'expected_value': '{}',
                    'name': 'join_types',
                    'raw_executed_value': {
                        'djblets_privacy_storedconsentdata': 'INNER JOIN',
                    },
                    'raw_expected_value': {},
                },
                {
                    'executed_value': 'Q(storedconsentdata__pk__gt=0)',
                    'expected_value': 'Q()',
                    'name': 'where',
                    'raw_executed_value': Q(storedconsentdata__pk__gt=0),
                    'raw_expected_value': Q(),
                },
            ])

    def test_with_mismatch_limit(self) -> None:
        """Testing compare_queries with mismatch and limit"""
        queries: ExpectedQueries = [
            {
                'model': User,
                'limit': 1,
            },
        ]

        with compare_queries(queries) as ctx:
            list(User.objects.all()[:2])

        self.assertTrue(ctx['has_mismatches'])
        self.assertEqual(ctx['num_executed_queries'], 1)
        self.assertEqual(ctx['num_expected_queries'], 1)
        self.assertFalse(ctx['query_count_mismatch'], 1)

        mismatches = ctx['query_mismatches']
        self.assertEqual(len(mismatches), 1)

        mismatch = mismatches[0]
        self.assertIsNotNone(mismatch['executed_query'])
        self.assertEqual(mismatch['index'], 0)
        self.assertIsNone(mismatch['note'])
        self.assertIsNotNone(mismatch['traceback'])
        self.assertEqual(
            mismatch['subqueries'],
            {
                '_unchecked_mismatched_attrs': set(),
                'has_mismatches': False,
                'num_executed_queries': 0,
                'num_expected_queries': 0,
                'query_count_mismatch': False,
                'query_mismatches': [],
            })
        self.assertEqual(
            mismatch['mismatched_attrs'],
            [
                {
                    'executed_value': '2',
                    'expected_value': '1',
                    'name': 'limit',
                    'raw_executed_value': 2,
                    'raw_expected_value': 1,
                },
            ])

    def test_with_mismatch_offset(self) -> None:
        """Testing compare_queries with mismatch and offset"""
        queries: ExpectedQueries = [
            {
                'model': User,
                'offset': 2,
            },
        ]

        with compare_queries(queries) as ctx:
            list(User.objects.all()[1:])

        self.assertTrue(ctx['has_mismatches'])
        self.assertEqual(ctx['num_executed_queries'], 1)
        self.assertEqual(ctx['num_expected_queries'], 1)
        self.assertFalse(ctx['query_count_mismatch'], 1)

        mismatches = ctx['query_mismatches']
        self.assertEqual(len(mismatches), 1)

        mismatch = mismatches[0]
        self.assertIsNotNone(mismatch['executed_query'])
        self.assertEqual(mismatch['index'], 0)
        self.assertIsNone(mismatch['note'])
        self.assertIsNotNone(mismatch['traceback'])
        self.assertEqual(
            mismatch['subqueries'],
            {
                '_unchecked_mismatched_attrs': set(),
                'has_mismatches': False,
                'num_executed_queries': 0,
                'num_expected_queries': 0,
                'query_count_mismatch': False,
                'query_mismatches': [],
            })
        self.assertEqual(
            mismatch['mismatched_attrs'],
            [
                {
                    'executed_value': '1',
                    'expected_value': '2',
                    'name': 'offset',
                    'raw_executed_value': 1,
                    'raw_expected_value': 2,
                },
            ])

    def test_with_mismatch_only_fields(self) -> None:
        """Testing compare_queries with mismatch and only_field"""
        queries: ExpectedQueries = [
            {
                'model': User,
                'only_fields': {'id'},
            },
        ]

        with compare_queries(queries) as ctx:
            list(User.objects.only('pk', 'username'))

        self.assertTrue(ctx['has_mismatches'])
        self.assertEqual(ctx['num_executed_queries'], 1)
        self.assertEqual(ctx['num_expected_queries'], 1)
        self.assertFalse(ctx['query_count_mismatch'], 1)

        mismatches = ctx['query_mismatches']
        self.assertEqual(len(mismatches), 1)

        mismatch = mismatches[0]
        self.assertIsNotNone(mismatch['executed_query'])
        self.assertEqual(mismatch['index'], 0)
        self.assertIsNone(mismatch['note'])
        self.assertIsNotNone(mismatch['traceback'])
        self.assertEqual(
            mismatch['subqueries'],
            {
                '_unchecked_mismatched_attrs': set(),
                'has_mismatches': False,
                'num_executed_queries': 0,
                'num_expected_queries': 0,
                'query_count_mismatch': False,
                'query_mismatches': [],
            })
        self.assertEqual(
            mismatch['mismatched_attrs'],
            [
                {
                    'executed_value': "{'id', 'username'}",
                    'expected_value': "{'id'}",
                    'name': 'only_fields',
                    'raw_executed_value': {'id', 'username'},
                    'raw_expected_value': {'id'},
                },
            ])

    def test_with_mismatch_order_by(self) -> None:
        """Testing compare_queries with mismatch and order_by"""
        queries: ExpectedQueries = [
            {
                'model': User,
                'order_by': ('pk',),
            },
        ]

        with compare_queries(queries) as ctx:
            list(User.objects.order_by('email'))

        self.assertTrue(ctx['has_mismatches'])
        self.assertEqual(ctx['num_executed_queries'], 1)
        self.assertEqual(ctx['num_expected_queries'], 1)
        self.assertFalse(ctx['query_count_mismatch'], 1)

        mismatches = ctx['query_mismatches']
        self.assertEqual(len(mismatches), 1)

        mismatch = mismatches[0]
        self.assertIsNotNone(mismatch['executed_query'])
        self.assertEqual(mismatch['index'], 0)
        self.assertIsNone(mismatch['note'])
        self.assertIsNotNone(mismatch['traceback'])
        self.assertEqual(
            mismatch['subqueries'],
            {
                '_unchecked_mismatched_attrs': set(),
                'has_mismatches': False,
                'num_executed_queries': 0,
                'num_expected_queries': 0,
                'query_count_mismatch': False,
                'query_mismatches': [],
            })
        self.assertEqual(
            mismatch['mismatched_attrs'],
            [
                {
                    'executed_value': "('email',)",
                    'expected_value': "('pk',)",
                    'name': 'order_by',
                    'raw_executed_value': ('email',),
                    'raw_expected_value': ('pk',),
                },
            ])

    def test_with_mismatch_select_for_update(self) -> None:
        """Testing compare_queries with mismatch and select_for_update"""
        queries: ExpectedQueries = [
            {
                'model': User,
            },
        ]

        with compare_queries(queries) as ctx:
            list(User.objects.select_for_update())

        self.assertTrue(ctx['has_mismatches'])
        self.assertEqual(ctx['num_executed_queries'], 1)
        self.assertEqual(ctx['num_expected_queries'], 1)
        self.assertFalse(ctx['query_count_mismatch'], 1)

        mismatches = ctx['query_mismatches']
        self.assertEqual(len(mismatches), 1)

        mismatch = mismatches[0]
        self.assertIsNotNone(mismatch['executed_query'])
        self.assertEqual(mismatch['index'], 0)
        self.assertIsNone(mismatch['note'])
        self.assertIsNotNone(mismatch['traceback'])
        self.assertEqual(
            mismatch['subqueries'],
            {
                '_unchecked_mismatched_attrs': set(),
                'has_mismatches': False,
                'num_executed_queries': 0,
                'num_expected_queries': 0,
                'query_count_mismatch': False,
                'query_mismatches': [],
            })
        self.assertEqual(
            mismatch['mismatched_attrs'],
            [
                {
                    'executed_value': 'True',
                    'expected_value': 'False',
                    'name': 'select_for_update',
                    'raw_executed_value': True,
                    'raw_expected_value': False,
                },
            ])

    def test_with_mismatch_select_related(self) -> None:
        """Testing compare_queries with mismatch and select_related"""
        queries: ExpectedQueries = [
            {
                'model': User,
                'select_related': {'group'},
            },
        ]

        with compare_queries(queries) as ctx:
            list(User.objects.select_related('storedconsentdata'))

        self.assertTrue(ctx['has_mismatches'])
        self.assertEqual(ctx['num_executed_queries'], 1)
        self.assertEqual(ctx['num_expected_queries'], 1)
        self.assertFalse(ctx['query_count_mismatch'], 1)

        mismatches = ctx['query_mismatches']
        self.assertEqual(len(mismatches), 1)

        mismatch = mismatches[0]
        self.assertIsNotNone(mismatch['executed_query'])
        self.assertEqual(mismatch['index'], 0)
        self.assertIsNone(mismatch['note'])
        self.assertIsNotNone(mismatch['traceback'])
        self.assertEqual(
            mismatch['subqueries'],
            {
                '_unchecked_mismatched_attrs': set(),
                'has_mismatches': False,
                'num_executed_queries': 0,
                'num_expected_queries': 0,
                'query_count_mismatch': False,
                'query_mismatches': [],
            })
        self.assertEqual(
            mismatch['mismatched_attrs'],
            [
                {
                    'executed_value': "{'storedconsentdata'}",
                    'expected_value': "{'group'}",
                    'name': 'select_related',
                    'raw_executed_value': {'storedconsentdata'},
                    'raw_expected_value': {'group'},
                },
            ])

    def test_with_mismatch_subqueries(self) -> None:
        """Testing compare_queries with mismatch and subqueries"""
        queries: ExpectedQueries = [
            {
                'model': User,
                'subqueries': [
                    {
                        'model': User,
                        'subqueries': [
                            {
                                'model': User,
                            },
                        ],
                    },
                ],
            },
        ]

        with compare_queries(queries) as ctx:
            count = (
                User.objects
                .filter(Exists(User.objects.filter(pk=OuterRef('pk'))))
                .distinct()
                .count()
            )

        self.assertEqual(count, 0)

        self.assertTrue(ctx['has_mismatches'])
        self.assertEqual(ctx['num_executed_queries'], 1)
        self.assertEqual(ctx['num_expected_queries'], 1)
        self.assertFalse(ctx['query_count_mismatch'], 1)

        mismatches = ctx['query_mismatches']
        self.assertEqual(len(mismatches), 1)

        mismatch = mismatches[0]
        self.assertIsNotNone(mismatch['executed_query'])
        self.assertEqual(mismatch['index'], 0)
        self.assertIsNone(mismatch['note'])
        self.assertIsNotNone(mismatch['traceback'])
        self.assertEqual(
            mismatch['mismatched_attrs'],
            [
                {
                    'executed_value': "{'__count': Count('*')}",
                    'expected_value': '{}',
                    'name': 'annotations',
                    'raw_executed_value': {'__count': Count('*')},
                    'raw_expected_value': {},
                },
            ])

        # Check the distinct subquery.
        subqueries = mismatch['subqueries']
        assert subqueries is not None
        self.assertTrue(subqueries['has_mismatches'])
        self.assertEqual(subqueries['num_executed_queries'], 1)
        self.assertEqual(subqueries['num_expected_queries'], 1)
        self.assertFalse(subqueries['query_count_mismatch'])

        subquery_mismatches = subqueries['query_mismatches']
        self.assertEqual(len(subquery_mismatches), 1)

        subquery_mismatch = subquery_mismatches[0]
        self.assertIsNotNone(subquery_mismatch['executed_query'])
        self.assertEqual(subquery_mismatch['index'], 0)
        self.assertIsNone(subquery_mismatch['note'])
        self.assertIsNone(subquery_mismatch['traceback'])
        self.assertEqual(
            subquery_mismatch['mismatched_attrs'],
            [
                {
                    'executed_value': 'True',
                    'expected_value': 'False',
                    'name': 'distinct',
                    'raw_executed_value': True,
                    'raw_expected_value': False,
                },
                {
                    'executed_value': 'True',
                    'expected_value': 'False',
                    'name': 'subquery',
                    'raw_executed_value': True,
                    'raw_expected_value': False,
                },
                {
                    'executed_value': 'Q(__Exists__subquery__=1)',
                    'expected_value': 'Q()',
                    'name': 'where',
                    'raw_executed_value': Q(__Exists__subquery__=1),
                    'raw_expected_value': Q(),
                },
            ])

        # Check the Exists subquery.
        subqueries = subquery_mismatch['subqueries']
        assert subqueries is not None
        self.assertTrue(subqueries['has_mismatches'])
        self.assertEqual(subqueries['num_executed_queries'], 1)
        self.assertEqual(subqueries['num_expected_queries'], 1)
        self.assertFalse(subqueries['query_count_mismatch'])

        subquery_mismatches = subqueries['query_mismatches']
        self.assertEqual(len(subquery_mismatches), 1)

        subquery_mismatch = subquery_mismatches[0]
        self.assertIsNotNone(subquery_mismatch['executed_query'])
        self.assertEqual(subquery_mismatch['index'], 0)
        self.assertIsNone(subquery_mismatch['note'])
        self.assertIsNone(subquery_mismatch['traceback'])
        self.assertEqual(
            subquery_mismatch['mismatched_attrs'],
            [
                {
                    'executed_value': "{'a': Value(1)}",
                    'expected_value': '{}',
                    'name': 'annotations',
                    'raw_executed_value': {'a': Value(1)},
                    'raw_expected_value': {},
                },
                {
                    'executed_value': 'True',
                    'expected_value': 'False',
                    'name': 'subquery',
                    'raw_executed_value': True,
                    'raw_expected_value': False,
                },
                {
                    'executed_value': '1',
                    'expected_value': 'None',
                    'name': 'limit',
                    'raw_executed_value': 1,
                    'raw_expected_value': None,
                },
                {
                    'executed_value': "Q(pk=OuterRef('pk'))",
                    'expected_value': 'Q()',
                    'name': 'where',
                    'raw_executed_value': Q(pk=OuterRef('pk')),
                    'raw_expected_value': Q(),
                },
            ])
        self.assertEqual(
            subquery_mismatch['subqueries'],
            {
                '_unchecked_mismatched_attrs': set(),
                'has_mismatches': False,
                'num_executed_queries': 0,
                'num_expected_queries': 0,
                'query_count_mismatch': False,
                'query_mismatches': [],
            })

    def test_with_mismatch_type(self) -> None:
        """Testing compare_queries with mismatch and type"""
        queries: ExpectedQueries = [
            {
                'model': User,
            },
        ]

        with compare_queries(queries) as ctx:
            User.objects.create(username='test-user')

        self.assertTrue(ctx['has_mismatches'])
        self.assertEqual(ctx['num_executed_queries'], 1)
        self.assertEqual(ctx['num_expected_queries'], 1)
        self.assertFalse(ctx['query_count_mismatch'], 1)

        mismatches = ctx['query_mismatches']
        self.assertEqual(len(mismatches), 1)

        mismatch = mismatches[0]
        self.assertIsNotNone(mismatch['executed_query'])
        self.assertEqual(mismatch['index'], 0)
        self.assertIsNone(mismatch['note'])
        self.assertIsNotNone(mismatch['traceback'])
        self.assertEqual(
            mismatch['subqueries'],
            {
                '_unchecked_mismatched_attrs': set(),
                'has_mismatches': False,
                'num_executed_queries': 0,
                'num_expected_queries': 0,
                'query_count_mismatch': False,
                'query_mismatches': [],
            })
        self.assertEqual(
            mismatch['mismatched_attrs'],
            [
                {
                    'executed_value': "'INSERT'",
                    'expected_value': "'SELECT'",
                    'name': 'type',
                    'raw_executed_value': 'INSERT',
                    'raw_expected_value': 'SELECT',
                },
            ])

    def test_with_mismatch_values_select(self) -> None:
        """Testing compare_queries with mismatch and values_select"""
        queries: ExpectedQueries = [
            {
                'model': User,
                'values_select': ('pk',),
            },
        ]

        with compare_queries(queries) as ctx:
            list(User.objects.values('username', 'email'))

        self.assertTrue(ctx['has_mismatches'])
        self.assertEqual(ctx['num_executed_queries'], 1)
        self.assertEqual(ctx['num_expected_queries'], 1)
        self.assertFalse(ctx['query_count_mismatch'], 1)

        mismatches = ctx['query_mismatches']
        self.assertEqual(len(mismatches), 1)

        mismatch = mismatches[0]
        self.assertIsNotNone(mismatch['executed_query'])
        self.assertEqual(mismatch['index'], 0)
        self.assertIsNone(mismatch['note'])
        self.assertIsNotNone(mismatch['traceback'])
        self.assertEqual(
            mismatch['subqueries'],
            {
                '_unchecked_mismatched_attrs': set(),
                'has_mismatches': False,
                'num_executed_queries': 0,
                'num_expected_queries': 0,
                'query_count_mismatch': False,
                'query_mismatches': [],
            })
        self.assertEqual(
            mismatch['mismatched_attrs'],
            [
                {
                    'executed_value': "('username', 'email')",
                    'expected_value': "('pk',)",
                    'name': 'values_select',
                    'raw_executed_value': ('username', 'email'),
                    'raw_expected_value': ('pk',),
                },
            ])

    def test_with_mismach_where(self) -> None:
        """Testing compare_queries with mismatch and where"""
        queries: ExpectedQueries = [
            {
                'model': User,
                'where': Q(username='test'),
            },
        ]

        with compare_queries(queries) as ctx:
            list(User.objects.filter(is_active=True))

        self.assertTrue(ctx['has_mismatches'])
        self.assertEqual(ctx['num_executed_queries'], 1)
        self.assertEqual(ctx['num_expected_queries'], 1)
        self.assertFalse(ctx['query_count_mismatch'], 1)

        mismatches = ctx['query_mismatches']
        self.assertEqual(len(mismatches), 1)

        mismatch = mismatches[0]
        self.assertIsNotNone(mismatch['executed_query'])
        self.assertEqual(mismatch['index'], 0)
        self.assertIsNone(mismatch['note'])
        self.assertIsNotNone(mismatch['traceback'])
        self.assertEqual(
            mismatch['subqueries'],
            {
                '_unchecked_mismatched_attrs': set(),
                'has_mismatches': False,
                'num_executed_queries': 0,
                'num_expected_queries': 0,
                'query_count_mismatch': False,
                'query_mismatches': [],
            })
        self.assertEqual(
            mismatch['mismatched_attrs'],
            [
                {
                    'executed_value': 'Q(is_active=True)',
                    'expected_value': "Q(username='test')",
                    'name': 'where',
                    'raw_executed_value': Q(is_active=True),
                    'raw_expected_value': Q(username='test'),
                },
            ])

    def test_with_note(self) -> None:
        """Testing compare_queries with mismatch and note"""
        queries: ExpectedQueries = [
            {
                'model': Group,
            },
        ]

        with compare_queries(queries) as ctx:
            list(User.objects.filter(is_active=True))

        self.assertTrue(ctx['has_mismatches'])
        self.assertEqual(ctx['num_executed_queries'], 1)
        self.assertEqual(ctx['num_expected_queries'], 1)
        self.assertFalse(ctx['query_count_mismatch'], 1)

        mismatches = ctx['query_mismatches']
        self.assertEqual(len(mismatches), 1)

        mismatch = mismatches[0]
        self.assertIsNotNone(mismatch['executed_query'])
        self.assertEqual(mismatch['index'], 0)
        self.assertIsNone(mismatch['note'])
        self.assertIsNotNone(mismatch['traceback'])
        self.assertEqual(
            mismatch['subqueries'],
            {
                '_unchecked_mismatched_attrs': set(),
                'has_mismatches': False,
                'num_executed_queries': 0,
                'num_expected_queries': 0,
                'query_count_mismatch': False,
                'query_mismatches': [],
            })
        self.assertEqual(
            mismatch['mismatched_attrs'],
            [
                {
                    'executed_value': repr(User),
                    'expected_value': repr(Group),
                    'name': 'model',
                    'raw_executed_value': User,
                    'raw_expected_value': Group,
                },
                {
                    'executed_value': "{'auth_user'}",
                    'expected_value': "{'auth_group'}",
                    'name': 'tables',
                    'raw_executed_value': {'auth_user'},
                    'raw_expected_value': {'auth_group'},
                },
                {
                    'executed_value': 'Q(is_active=True)',
                    'expected_value': 'Q()',
                    'name': 'where',
                    'raw_executed_value': Q(is_active=True),
                    'raw_expected_value': Q(),
                },
            ])


class FormatNodeTests(TestCase):
    """Unit tests for djblets.db.query_comparator._format_node.

    Version Added:
        3.4
    """

    ######################
    # Instance variables #
    ######################

    catch_ctx: CatchQueriesContext

    def setUp(self) -> None:
        super().setUp()

        self.catch_ctx = CatchQueriesContext(executed_queries=[],
                                             queries_to_qs={})

    def tearDown(self) -> None:
        self.catch_ctx = None  # type: ignore

        super().tearDown()

    def test_with_nested_q(self) -> None:
        """Testing _format_node with single nested Qs"""
        self.assertEqual(
            _format_node(_normalize_q(Q(a=1)),
                         catch_ctx=self.catch_ctx),
            'Q(a=1)')

    def test_with_nested_q_negated_start(self) -> None:
        """Testing _format_node with single nested Qs and negated at start"""
        self.assertEqual(
            _format_node(_normalize_q(~Q(a=1)),
                         catch_ctx=self.catch_ctx),
            '~Q(a=1)')

    def test_with_nested_q_negated_start_2(self) -> None:
        """Testing _format_node with single nested Qs and negated at start"""
        self.assertEqual(
            _format_node(_normalize_q(~Q(group__in=[1])),
                         catch_ctx=self.catch_ctx),
            '~Q(group__in=[1])')

    def test_with_nested_q_negated_all(self) -> None:
        """Testing _format_node with single nested Qs and all negated states.
        """
        print(_normalize_q(~Q(~Q(~Q(~Q(a=1))))))
        self.assertEqual(
            _format_node(_normalize_q(~Q(~Q(~Q(~Q(a=1))))),
                         catch_ctx=self.catch_ctx),
            '~Q(\n'
            ' ~Q(\n'
            '  ~Q(\n'
            '   ~Q(a=1))))')

    def test_with_negated_with_n_children_and(self) -> None:
        """Testing _format_node with negated Q with nested children with AND"""
        self.assertEqual(
            _format_node(_normalize_q(~Q(Q(a=1) & Q(b=2))),
                         catch_ctx=self.catch_ctx),
            '~Q(\n'
            '   Q(a=1)\n'
            ' & Q(b=2))')

    def test_with_negated_with_n_children_or(self) -> None:
        """Testing _format_node with negated Q with nested children with OR"""
        self.assertEqual(
            _format_node(_normalize_q(~Q(Q(a=1) | Q(b=2))),
                         catch_ctx=self.catch_ctx),
            '~Q(\n'
            '   Q(a=1)\n'
            ' | Q(b=2))')

    def test_with_negated_with_subquery(self) -> None:
        """Testing _format_node with negated Q with nested subquery"""
        self.assertEqual(
            _format_node(_normalize_q(~Q(Exists(User.objects.all()))),
                         catch_ctx=self.catch_ctx),
            '~Q(__Exists__subquery__=1)')

    def test_with_subquery(self) -> None:
        """Testing _format_node with subquery"""
        self.assertEqual(
            _format_node(
                _normalize_q(
                    ~Q(Exists(User.objects.all()) |
                       ~Exists(User.objects.none()))),
                catch_ctx=self.catch_ctx),
            '~Q(\n'
            '   Q(__Exists__subquery__=1)\n'
            ' | ~Q(__Exists__subquery__=2))')

    def test_with_subquery_nested(self) -> None:
        """Testing _format_node with subquery and nested expressions"""
        self.assertEqual(
            _format_node(
                _normalize_q(
                    Exists(User.objects.filter(
                        Q(is_active=True) &
                        ~Q(username='test'))) |
                    ~Exists(User.objects.filter(
                        Q(email__contains='@hotmail.com') &
                        ~Q(first_name='Bob')))),
                catch_ctx=self.catch_ctx),
            'Q(\n'
            '   Q(__Exists__subquery__=1)\n'
            ' | ~Q(__Exists__subquery__=2))')

    def test_with_f(self) -> None:
        """Testing _format_node with F() expression"""
        self.assertEqual(
            _format_node(
                _normalize_q(Q(a=F('b') + 3)),
                catch_ctx=self.catch_ctx),
            "Q(a=F('b') + 3)")

    def test_with_complex(self) -> None:
        """Testing _format_node with complex nested tree of Qs and Subqueries
        """
        self.assertEqual(
            _format_node(
                _normalize_q(
                    Q(Q(Q(a=1) &
                        (Q(Q(b=2)) |
                         Q(c=3) | Q())) |
                      ~Q(Q(Q(d=4)) |
                         (Q(Exists(User.objects.all())) &
                          ~Exists(User.objects.none()) &
                          Q(e=5) &
                          Q() &
                          ~Q() &
                          ~Q(Q(Q(f=F('g') + 6))))))),
                catch_ctx=self.catch_ctx),
            "Q(\n"
            "   Q(\n"
            "      Q(a=1)\n"
            "    & Q(\n"
            "         Q(b=2)\n"
            "       | Q(c=3)))\n"
            " | ~Q(\n"
            "      Q(d=4)\n"
            "    | Q(\n"
            "         Q(__Exists__subquery__=1)\n"
            "       & ~Q(__Exists__subquery__=2)\n"
            "       & Q(e=5)\n"
            "       & ~Q(f=F('g') + 6))))")


class NormalizeQTests(TestCase):
    """Unit tests for djblets.db.query_comparator._normalize_q.

    Version Added:
        3.4
    """

    def test_with_nested_q(self) -> None:
        """Testing _normalize_q with single nested Qs"""
        norm_q = _normalize_q(Q(Q(Q(Q(a=1)))))
        self.assertEqual(norm_q, Q(a=1))
        self.assertEqual(repr(norm_q), "<Q: (AND: ('a', 1))>")

    def test_with_nested_q_and_empty(self) -> None:
        """Testing _normalize_q with nested Qs and empty Qs"""
        norm_q = _normalize_q(Q(Q(Q(Q(a=1) & Q()))))
        self.assertEqual(norm_q, Q(a=1))
        self.assertEqual(repr(norm_q), "<Q: (AND: ('a', 1))>")

    def test_with_nested_q_negated_start(self) -> None:
        """Testing _normalize_q with single nested Qs and negated at start"""
        norm_q = _normalize_q(~Q(Q(Q(Q(a=1)))))
        self.assertEqual(norm_q, ~Q(a=1))
        self.assertEqual(repr(norm_q), "<Q: (NOT (AND: ('a', 1)))>")

    def test_with_nested_q_negated_start_2(self) -> None:
        """Testing _normalize_q with single nested Qs and negated at start"""
        norm_q = _normalize_q(~Q(Q(group__in=[1])))
        self.assertEqual(norm_q, ~Q(group__in=[1]))
        self.assertEqual(repr(norm_q), "<Q: (NOT (AND: ('group__in', [1])))>")

    def test_with_nested_q_negated_middle(self) -> None:
        """Testing _normalize_q with single nested Qs and negated in the middle
        """
        norm_q = _normalize_q(Q(Q(~Q(Q(a=1)))))
        self.assertEqual(norm_q, ~Q(a=1))
        self.assertEqual(repr(norm_q), "<Q: (NOT (AND: ('a', 1)))>")

    def test_with_nested_q_negated_end(self) -> None:
        """Testing _normalize_q with single nested Qs and negated at the end
        """
        norm_q = _normalize_q(Q(Q(Q(~Q(a=1)))))
        self.assertEqual(norm_q, ~Q(a=1))
        self.assertEqual(repr(norm_q), "<Q: (NOT (AND: ('a', 1)))>")

    def test_with_nested_q_negated_all(self) -> None:
        """Testing _normalize_q with single nested Qs and all negated states.
        """
        norm_q = _normalize_q(~Q(~Q(~Q(~Q(a=1)))))
        self.assertEqual(norm_q, ~Q(~Q(~Q(~Q(a=1)))))
        self.assertEqual(
            repr(norm_q),
            "<Q: (NOT (AND: (NOT (AND: (NOT (AND: (NOT (AND: ('a', 1)))))))))>"
        )

    def test_with_nested_q_negated_double(self) -> None:
        """Testing _normalize_q with single nested Qs and double-negated"""
        norm_q = _normalize_q(~Q(Q(Q(~Q(a=1)))))
        self.assertEqual(norm_q, ~Q(~Q(a=1)))
        self.assertEqual(repr(norm_q),
                         "<Q: (NOT (AND: (NOT (AND: ('a', 1)))))>")

    def test_with_negated_with_n_children_and(self) -> None:
        """Testing _normalize_q with negated Q with nested children with AND"""
        norm_q = _normalize_q(~Q(Q(a=1) & Q(b=2)))
        self.assertEqual(norm_q, ~Q(Q(a=1) & Q(b=2)))
        self.assertEqual(repr(norm_q),
                         "<Q: (NOT (AND: (AND: ('a', 1), ('b', 2))))>")

    def test_with_negated_with_n_children_or(self) -> None:
        """Testing _normalize_q with negated Q with nested children with OR"""
        norm_q = _normalize_q(~Q(Q(a=1) | Q(b=2)))
        self.assertEqual(norm_q, ~Q(Q(a=1) | Q(b=2)))
        self.assertEqual(repr(norm_q),
                         "<Q: (NOT (AND: (OR: ('a', 1), ('b', 2))))>")

    def test_with_negated_with_subquery(self) -> None:
        """Testing _normalize_q with negated Q with nested subquery"""
        norm_q = _normalize_q(~Q(Exists(User.objects.all())))
        self.assertEqual(norm_q, ~Q(__Exists__subquery__=1))
        self.assertEqual(repr(norm_q),
                         "<Q: (NOT (AND: ('__Exists__subquery__', 1)))>")

    def test_with_subquery(self) -> None:
        """Testing _normalize_q with subquery"""
        norm_q = _normalize_q(~Q(Exists(User.objects.all()) |
                                 ~Exists(User.objects.none())))
        self.assertEqual(norm_q,
                         ~Q(Q(__Exists__subquery__=1) |
                            ~Q(__Exists__subquery__=2)))
        self.assertEqual(
            repr(norm_q),
            "<Q: (NOT (AND: (OR: ('__Exists__subquery__', 1),"
            " (NOT (AND: ('__Exists__subquery__', 2))))))>")

    def test_with_subquery_nested(self) -> None:
        """Testing _normalize_q with subquery and nested expressions"""
        norm_q = _normalize_q(
            Exists(User.objects.filter(
                Q(is_active=True) &
                ~Q(username='test')
            )) |
            ~Exists(User.objects.filter(
                Q(email__contains='@hotmail.com') &
                ~Q(first_name='Bob')
            )))
        self.assertEqual(norm_q,
                         (Q(__Exists__subquery__=1) |
                          ~Q(__Exists__subquery__=2)))
        self.assertEqual(
            repr(norm_q),
            "<Q: (OR: ('__Exists__subquery__', 1),"
            " (NOT (AND: ('__Exists__subquery__', 2))))>")

    def test_with_complex(self) -> None:
        """Testing _normalize_q with complex nested tree of Qs and Subqueries
        """
        norm_q = _normalize_q(
            Q(Q(Q(a=1) & (Q(Q(b=2)) | Q(c=3) | Q())) |
              ~Q(Q(Q(d=4)) |
                 (Q(Exists(User.objects.all())) &
                  ~Exists(User.objects.none()) &
                  Q(e=5) &
                  Q() &
                  ~Q() &
                  ~Q(Q(f=6))))))
        self.assertEqual(
            norm_q,
            Q(Q(a=1) &
              Q(Q(b=2) | Q(c=3))) |
            ~Q(Q(d=4) |
               (Q(__Exists__subquery__=1) &
                ~Q(__Exists__subquery__=2) &
                (Q(e=5) & ~Q(f=6)))))
        self.assertEqual(
            repr(norm_q),
            "<Q: (OR: (AND: ('a', 1), (OR: ('b', 2), ('c', 3))),"
            " (NOT (AND: (OR: ('d', 4), (AND: ('__Exists__subquery__', 1),"
            " (NOT (AND: ('__Exists__subquery__', 2))),"
            " ('e', 5), (NOT (AND: ('f', 6))))))))>")



"""A cache backend that forwards to other dynamically-configured backends."""

DEFAULT_FORWARD_CACHE_ALIAS = 'forwarded_backend'


class ForwardingCacheBackend(object):
    """Forwards requests to another cache backend.

    This is used to allow for dynamic configuration of caches that can be
    swapped in and out. By setting this cache backend as the default backend,
    the consumer can easily switch between other cache backends without
    modifying :file:`settings.py` and restarting the app.

    This by default looks for another cache backend in
    ``settings.CACHES['forwarded_backend']``. This can be changed with the
    ``LOCATION`` setting for this cache backend. All requests and attribute
    lookups will be forwarded there.

    If a consumer switches the real cache backend, it can call
    :py:meth:`reset_backend`, and all future cache requests will go to the
    newly computed backend.
    """

    def __init__(self, cache_name=DEFAULT_FORWARD_CACHE_ALIAS,
                 *args, **kwargs):
        self._cache_name = cache_name

    @property
    def backend(self):
        """Return the forwarded cache backend."""
        from django.core.cache import caches

        return caches[self._cache_name]

    def reset_backend(self):
        """Reset the forwarded cache backend.

        This must be called after modifying
        ``settings.CACHES['forwarded_backend']`` in order for the new
        backend to be picked up.
        """
        from django.core.cache import caches

        try:
            del caches[self._cache_name]
        except Exception:
            # We don't really care if this fails. It probably means the
            # configuration was already deleted. The next access will recreate
            # it.
            pass

    def __contains__(self, key):
        return key in self.backend

    def __getattribute__(self, name):
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            return self.backend.__getattribute__(name)



"""An authentication backend for HTTP Basic Auth."""

from __future__ import annotations

import base64
import logging
from typing import List

from django.http import HttpRequest

from djblets.webapi.auth.backends.base import (WebAPIAuthBackend,
                                               WebAPIGetCredentialsResult)


logger = logging.getLogger(__name__)


class WebAPIBasicAuthBackend(WebAPIAuthBackend):
    """Handles HTTP Basic Authentication for the web API."""

    www_auth_scheme = 'Basic realm="Web API"'

    def get_credentials(
        self,
        request: HttpRequest,
    ) -> WebAPIGetCredentialsResult:
        """Return credentials provided in the request.

        This returns a dictionary of all credentials necessary for this
        auth backend.

        If an :mailheader:`Authorization` header is provided, the result will
        be the parsed username and password.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            dict:
            A dictionary of credential information containing:

            Keys:
                password (str):
                    The password from the authentication request.

                username (str):
                    The username from the authentication request.

            This will be ``None`` if :mailheader:`Authorization` is not passed
            or could not be parsed as a HTTP Basic Auth header.
        """
        parts: List[str] = request.META['HTTP_AUTHORIZATION'].split(' ')
        realm = parts[0]

        if realm != 'Basic':
            return None

        try:
            encoded_auth: bytes = parts[1].encode('utf-8')
            username: str
            password: str

            username, password = \
                base64.b64decode(encoded_auth).decode('utf-8').split(':', 1)
        except Exception:
            logger.warning('Failed to parse HTTP_AUTHORIZATION header %s',
                           request.META['HTTP_AUTHORIZATION'],
                           exc_info=True,
                           extra={'request': request})
            return None

        return {
            'username': username,
            'password': password,
        }



import io
from html.entities import codepoint2name

from markdown import __version_info__ as markdown_version, markdown

from djblets.testing.testcases import TestCase
from djblets.markdown import (get_markdown_element_tree,
                              iter_markdown_lines,
                              markdown_escape,
                              markdown_unescape,
                              render_markdown_from_file)


class MarkdownTestCase(TestCase):
    """Base class for Markdown-related test cases."""

    def render_markdown(self, text):
        """Render Markdown text with a standard configuration.

        Args:
            text (unicode):
                The Markdown-formatted text to render.

        Returns:
            unicode:
            The resulting HTML.
        """
        return markdown(
            text,
            output_format='xhtml1',
            lazy_ol=False,
            extensions=[
                'markdown.extensions.fenced_code',
                'markdown.extensions.codehilite',
                'markdown.extensions.sane_lists',
                'markdown.extensions.nl2br',
                'djblets.markdown.extensions.escape_html',
                'djblets.markdown.extensions.wysiwyg',
            ],
            extension_configs={
                'markdown.extensions.codehilite': {
                    'guess_lang': False,
                },
            },
        )


class MarkdownUtilsTests(MarkdownTestCase):
    """Unit tests for the Markdown utility functions."""

    UNESCAPED_TEXT = r'\`*_{}[]()#+-.!'
    ESCAPED_TEXT = r'\\\`\*\_\{\}\[\]\(\)#+-.\!'

    def test_get_markdown_element_tree(self):
        """Testing get_markdown_element_tree"""
        node = get_markdown_element_tree(
            self.render_markdown('**Test**\nHi.'))

        self.assertEqual(node[0].toxml(),
                         '<p><strong>Test</strong><br/>\n'
                         'Hi.</p>')

    def test_get_markdown_element_tree_with_illegal_chars(self):
        """Testing get_markdown_element_tree with illegal characters"""
        node = get_markdown_element_tree(
            self.render_markdown('(**Test**\x0C)'))

        self.assertEqual(node[0].toxml(), '<p>(<strong>Test</strong>)</p>')

    def test_get_markdown_element_tree_with_illegal_in_email(self):
        """Testing get_markdown_element_tree with illegal characters in
        email address
        """
        node = get_markdown_element_tree(
            self.render_markdown(r'<test@test.com\\>'))

        self.assertEqual(
            node[0].toxml(),
            '<p><a href="mailto:test@test.com?92?">test@test.com?92?</a></p>')

    def test_get_markdown_element_tree_with_named_entities(self):
        """Testing get_markdown_element_tree with named entities"""
        rendered_html_entities = ['&fooooo;']
        expected_html_entities = ['?']

        # toxml() will convert a select list of characters into named
        # entities when generating the string. We need to account for this.
        # These aren't expanded when text is processed in
        # get_markdown_element_tree(). They'll be &#...; entities.
        toxml_expanded_chars = {
            '&': '&amp;',
            '<': '&lt;',
            '>': '&gt;',
            '"': '&quot;',
        }

        for char_code, entity_name in codepoint2name.items():
            rendered_html_entities.append('&%s;' % entity_name)

            char = chr(char_code)
            expected_html_entities.append(toxml_expanded_chars.get(char, char))

        node = get_markdown_element_tree(''.join(rendered_html_entities))
        self.assertEqual(node[0].toxml(), ''.join(expected_html_entities))

    def test_markdown_escape(self):
        """Testing markdown_escape"""
        self.assertEqual(markdown_escape(self.UNESCAPED_TEXT),
                         self.ESCAPED_TEXT)

    def test_markdown_escape_periods(self):
        """Testing markdown_escape with '.' placement"""
        self.assertEqual(
            markdown_escape('Line. 1.\n'
                            '1. Line. 2.\n'
                            '1.2. Line. 3.\n'
                            '  1. Line. 4.'),
            ('Line. 1.\n'
             '1\\. Line. 2.\n'
             '1.2. Line. 3.\n'
             '  1\\. Line. 4.'))

    def test_markdown_escape_atx_headers(self):
        """Testing markdown_escape with '#' placement"""
        self.assertEqual(
            markdown_escape('### Header\n'
                            '  ## Header ##\n'
                            'Not # a header'),
            ('\\#\\#\\# Header\n'
             '  \\#\\# Header ##\n'
             'Not # a header'))

    def test_markdown_escape_hyphens(self):
        """Testing markdown_escape with '-' placement"""
        self.assertEqual(
            markdown_escape('Header\n'
                            '------\n'
                            '\n'
                            '- List item\n'
                            '  - List item\n'
                            'Just hyp-henated'),
            ('Header\n'
             '\\-\\-\\-\\-\\-\\-\n'
             '\n'
             '\\- List item\n'
             '  \\- List item\n'
             'Just hyp-henated'))

    def test_markdown_escape_plusses(self):
        """Testing markdown_escape with '+' placement"""
        self.assertEqual(
            markdown_escape('+ List item\n'
                            'a + b'),
            ('\\+ List item\n'
             'a + b'))

    def test_markdown_escape_underscores(self):
        """Testing markdown_escape with '_' placement"""
        self.assertEqual(markdown_escape('_foo_'), r'\_foo\_')
        self.assertEqual(markdown_escape('__foo__'), r'\_\_foo\_\_')
        self.assertEqual(markdown_escape(' _foo_ '), r' \_foo\_ ')
        self.assertEqual(markdown_escape('f_o_o'), r'f_o_o')
        self.assertEqual(markdown_escape('_f_o_o'), r'\_f_o_o')
        self.assertEqual(markdown_escape('f_o_o_'), r'f_o_o\_')
        self.assertEqual(markdown_escape('foo_ _bar'), r'foo\_ \_bar')
        self.assertEqual(markdown_escape('foo__bar'), r'foo__bar')
        self.assertEqual(markdown_escape('foo\n_bar'), 'foo\n\\_bar')
        self.assertEqual(markdown_escape('(_foo_)'), r'(\_foo\_)')

    def test_markdown_escape_asterisks(self):
        """Testing markdown_escape with '*' placement"""
        self.assertEqual(markdown_escape('*foo*'), r'\*foo\*')
        self.assertEqual(markdown_escape('**foo**'), r'\*\*foo\*\*')
        self.assertEqual(markdown_escape(' *foo* '), r' \*foo\* ')
        self.assertEqual(markdown_escape('f*o*o'), r'f*o*o')
        self.assertEqual(markdown_escape('f*o*o*'), r'f*o*o\*')
        self.assertEqual(markdown_escape('foo* *bar'), r'foo\* \*bar')
        self.assertEqual(markdown_escape('foo**bar'), r'foo**bar')
        self.assertEqual(markdown_escape('foo\n*bar'), 'foo\n\\*bar')

    def test_markdown_escape_parens(self):
        """Testing markdown_escape with '(' and ')' placement"""
        self.assertEqual(markdown_escape('[name](link)'), r'\[name\]\(link\)')
        self.assertEqual(markdown_escape('(link)'), r'(link)')
        self.assertEqual(markdown_escape('](link)'), r'\](link)')
        self.assertEqual(markdown_escape('[foo] ](link)'),
                         r'\[foo\] \](link)')

    def test_markdown_escape_gt_text(self):
        """Testing markdown_escape with '>' for standard text"""
        self.assertEqual(markdown_escape('<foo>'), r'<foo>')

    def test_markdown_escape_gt_blockquotes(self):
        """Testing markdown_escape with '>' for blockquotes"""
        self.assertEqual(markdown_escape('>'), r'\>')
        self.assertEqual(markdown_escape('> foo'), r'\> foo')
        self.assertEqual(markdown_escape('  > foo'), r'  \> foo')
        self.assertEqual(markdown_escape('> > foo'), r'\> \> foo')
        self.assertEqual(markdown_escape('  > > foo'), r'  \> \> foo')

    def test_markdown_escape_gt_autolinks(self):
        """Testing markdown_escape with '>' for autolinks"""
        self.assertEqual(markdown_escape('<http://www.example.com>'),
                         r'<http://www.example.com\>')

    def test_markdown_escape_gt_autoemail(self):
        """Testing markdown_escape with '>' for autoemails"""
        self.assertEqual(markdown_escape('<user@example.com>'),
                         r'<user@example.com\>')

    def test_markdown_unescape(self):
        """Testing markdown_unescape"""
        self.assertEqual(markdown_unescape(self.ESCAPED_TEXT),
                         self.UNESCAPED_TEXT)

        self.assertEqual(
            markdown_unescape('&nbsp;   code\n'
                              '&nbsp;   code'),
            ('    code\n'
             '    code'))
        self.assertEqual(
            markdown_unescape('&nbsp;\tcode\n'
                              '&nbsp;\tcode'),
            ('\tcode\n'
             '\tcode'))

    def test_markdown_list_start(self):
        """Testing iteration of Markdown lists with a 'start' parameter"""
        # In bug 3715, we were treating the 'start' parameter incorrectly. This
        # checks that said bug is fixed.
        rendered = '<ul start="2"><li>x</li><li>y</li></ul>'
        lines = list(iter_markdown_lines(rendered))

        self.assertEqual(lines, [
            '<ul start="2"><li>x</li></ul>',
            '<ul start="3"><li>y</li></ul>',
        ])

    def test_markdown_codehilite_blocks(self):
        """Testing iteration of Markdown codehilite blocks"""
        rendered = '\n'.join([
            '<div class="codehilite"><pre>Line one',
            'Line two',
            'Line three',
            'Line four',
            '</pre></div>',
        ])
        lines = list(iter_markdown_lines(rendered))

        self.assertEqual(lines, [
            ('<div class="codehilite codehilite-multiline-start"><pre>'
             'Line one</pre></div>'),
            ('<div class="codehilite codehilite-multiline-middle"><pre>'
             'Line two</pre></div>'),
            ('<div class="codehilite codehilite-multiline-middle"><pre>'
             'Line three</pre></div>'),
            ('<div class="codehilite codehilite-multiline-end"><pre>'
             'Line four</pre></div>'),
        ])

    def test_sanitize_illegal_chars(self):
        """Testing sanitize_illegal_chars_for_xml"""
        s = '<a>\u2018\u2019\u201c\u201d\u201c\u201d</a>'

        # This used to cause a UnicodeDecodeError
        nodes = get_markdown_element_tree(s)

        self.assertEqual(len(nodes), 1)
        self.assertEqual(nodes[0].toxml(),
                         '<a>\u2018\u2019\u201c\u201d\u201c\u201d</a>')

    def test_render_markdown_from_file(self):
        """Testing render_markdown_from_file"""
        stream = io.BytesIO(b'This is a **test**')

        self.assertEqual(render_markdown_from_file(stream),
                         '<p>This is a <strong>test</strong></p>')


class EscapeHtmlRenderTests(MarkdownTestCase):
    """Unit tests for rendering with the Escape HTML Markdown extension."""

    def test_render_escape(self):
        """Testing Markdown Escape HTML rendering with HTML content"""
        self.assertEqual(self.render_markdown('<b>Test</b>'),
                         '<p>&lt;b&gt;Test&lt;/b&gt;</p>')


class WysiwygRenderTests(MarkdownTestCase):
    """Unit tests for rendering with the WYSIWYG Markdown extension."""

    def test_code_1_blank_line(self):
        """Testing Markdown WYSIWYG rendering with code block and 1 surrounding
        blank line
        """
        self.assertEqual(
            self.render_markdown(
                'begin:\n'
                '\n'
                '    if (1) {}\n'
                '\n'
                'done.'),
            ('<p>begin:</p>\n'
             '%s'
             '<p>done.</p>')
            % self._build_expected_code_block('if (1) {}\n'))

    def test_code_2_blank_lines(self):
        """Testing Markdown WYSIWYG rendering with code block and 2 surrounding
        blank lines
        """
        self.assertEqual(
            self.render_markdown(
                'begin:\n'
                '\n'
                '\n'
                '    if (1) {}\n'
                '\n'
                '\n'
                'done.'),
            ('<p>begin:</p>\n'
             '<p></p>\n'
             '%s'
             '<p></p>\n'
             '<p>done.</p>')
            % self._build_expected_code_block('if (1) {}\n'))

    def test_code_3_blank_lines(self):
        """Testing Markdown WYSIWYG rendering with code block and 3 surrounding
        blank lines
        """
        self.assertEqual(
            self.render_markdown(
                'begin:\n'
                '\n'
                '\n'
                '\n'
                '    if (1) {}\n'
                '\n'
                '\n'
                '\n'
                'done.'),
            ('<p>begin:</p>\n'
             '<p></p>\n'
             '<p></p>\n'
             '%s'
             '<p></p>\n'
             '<p></p>\n'
             '<p>done.</p>')
            % self._build_expected_code_block('if (1) {}\n'))

    def test_code_4_blank_lines(self):
        """Testing Markdown WYSIWYG rendering with code block and 4 surrounding
        blank lines
        """
        self.assertEqual(
            self.render_markdown(
                'begin:\n'
                '\n'
                '\n'
                '\n'
                '\n'
                '    if (1) {}\n'
                '\n'
                '\n'
                '\n'
                '\n'
                'done.'),
            ('<p>begin:</p>\n'
             '<p></p>\n'
             '<p></p>\n'
             '<p></p>\n'
             '%s'
             '<p></p>\n'
             '<p></p>\n'
             '<p></p>\n'
             '<p>done.</p>')
            % self._build_expected_code_block('if (1) {}\n'))

    def test_lists_1_blank_line(self):
        """Testing Markdown WYSIWYG rendering with 1 blank lines between lists
        """
        # This really just results in a single list. This is Python Markdown
        # behavior.
        self.assertEqual(
            self.render_markdown(
                '1. item\n'
                '\n'
                '1. item'),
            ('<ol>\n'
             '<li>\n'
             '<p>item</p>\n'
             '</li>\n'
             '<li>\n'
             '<p>item</p>\n'
             '</li>\n'
             '</ol>'))

    def test_lists_2_blank_line(self):
        """Testing Markdown WYSIWYG rendering with 2 blank lines between lists
        """
        self.assertEqual(
            self.render_markdown(
                '1. item\n'
                '\n'
                '\n'
                '1. item'),
            ('<ol>\n'
             '<li>item</li>\n'
             '</ol>\n'
             '<p></p>\n'
             '<ol>\n'
             '<li>item</li>\n'
             '</ol>'))

    def test_lists_3_blank_line(self):
        """Testing Markdown WYSIWYG rendering with 3 blank lines between lists
        """
        self.assertEqual(
            self.render_markdown(
                '1. item\n'
                '\n'
                '\n'
                '\n'
                '1. item'),
            ('<ol>\n'
             '<li>item</li>\n'
             '</ol>\n'
             '<p></p>\n'
             '<p></p>\n'
             '<ol>\n'
             '<li>item</li>\n'
             '</ol>'))

    def test_ol(self):
        """Testing Markdown WYSIWYG rendering with ordered lists"""
        self.assertEqual(
            self.render_markdown(
                '1. Foo\n'
                '2. Bar'),
            ('<ol>\n'
             '<li>Foo</li>\n'
             '<li>Bar</li>\n'
             '</ol>'))

    def test_ol_start(self):
        """Testing Markdown WYSIWYG rendering with ordered lists using start=
        """
        self.assertEqual(
            self.render_markdown(
                '5. Foo\n'
                '6. Bar'),
            ('<ol start="5" style="counter-reset: li 4">\n'
             '<li>Foo</li>\n'
             '<li>Bar</li>\n'
             '</ol>'))

    def test_text_0_blank_lines(self):
        """Testing Markdown WYSIWYG rendering with 0 blank lines between text
        """
        self.assertEqual(
            self.render_markdown(
                'begin:\n'
                'done.'),
            ('<p>begin:<br />\n'
             'done.</p>'))

    def test_text_1_blank_line(self):
        """Testing Markdown WYSIWYG rendering with 1 blank line between text"""
        self.assertEqual(
            self.render_markdown(
                'begin:\n'
                '\n'
                'done.'),
            ('<p>begin:</p>\n'
             '<p>done.</p>'))

    def test_text_2_blank_lines(self):
        """Testing Markdown WYSIWYG rendering with 2 blank lines between text
        """
        self.assertEqual(
            self.render_markdown(
                'begin:\n'
                '\n'
                '\n'
                'done.'),
            ('<p>begin:</p>\n'
             '<p></p>\n'
             '<p>done.</p>'))

    def test_text_3_blank_lines(self):
        """Testing Markdown WYSIWYG rendering with 3 blank lines between text
        """
        self.assertEqual(
            self.render_markdown(
                'begin:\n'
                '\n'
                '\n'
                '\n'
                'done.'),
            ('<p>begin:</p>\n'
             '<p></p>\n'
             '<p></p>\n'
             '<p>done.</p>'))

    def test_trailing_p_trimmed(self):
        """Testing Markdown WYSIWYG rendering trims trailing paragraphs"""
        self.assertEqual(
            self.render_markdown(
                'begin:\n'
                '\n'
                '\n'),
            '<p>begin:</p>')

    def _build_expected_code_block(self, code):
        """Build the HTML that'd be expected for a Markdown code block.

        This handles the differences between Markdown < 3.2 code blocks
        (which will be used on Python 2.7) and Markdown >= 3.2 code blocks
        (Python 3 and up).

        Args:
            code (unicode):
                The code inside the code block.

        Returns:
            unicode:
            The HTML for the code block.
        """
        if markdown_version[:2] >= (3, 2):
            # Markdown 3.2+ adds <code> around any code blocks.
            code = '<code>%s</code>' % code

        return (
            '<div class="codehilite"><pre><span></span>%s</pre></div>\n'
            % code
        )



"""Specialized descriptors/properties for classes."""

from __future__ import annotations

import warnings
from typing import (Any, Callable, Generic, Optional, Sequence, Tuple, Type,
                    Union, cast, overload)

from typing_extensions import Self, TypeAlias, TypeVar


# NOTE: When mypy supports PEP 696, we can give many of these defaults.
#       Setters can be based on the preceding getters as specific in
#       Generic[...]. This will simplify usage for callers in the default
#       case.
_StoredT = TypeVar('_StoredT')
_GetT = TypeVar('_GetT')
_SetT = TypeVar('_SetT')

_AliasPropertySetT = TypeVar('_AliasPropertySetT')
_AliasPropertyStoredT: TypeAlias = Any

_TypedPropertyGetT = TypeVar('_TypedPropertyGetT')
_TypedPropertyValidTypesParamT: TypeAlias = Union[Type[_SetT],
                                                  Sequence[Type[_SetT]]]


class BaseProperty(Generic[_StoredT]):
    """Base class for a custom property for a class.

    This is an optional base class that provides handy utilities that
    properties may need. For instance, being able to determine the name of
    the property's attribute on a class.

    Version Changed:
        3.3:
        This now supports generics for typing, taking the type of the stored
        content.
    """

    ######################
    # Instance variables #
    ######################

    #: The name of the owning attribute.
    #:
    #: Version Added:
    #:     3.3
    attr_name: str

    def __set_name__(
        self,
        owner: type,
        name: str,
    ) -> None:
        """Handle setting the attribute name for this property.

        Version Added:
            3.3

        Args:
            owner (type, unused):
                The class that owns the property.

            name (str):
                The attribute name for this property.
        """
        self.attr_name = name


class AliasProperty(Generic[_GetT, _AliasPropertySetT],
                    BaseProperty[_AliasPropertyStoredT]):
    """A property that aliases to another property or attribute.

    Alias properties are used to automatically retrieve from another property
    on access, or to set a value on another property. It's useful when
    wanting to rename an attribute but continue to provide a deprecated
    version, or when creating an object that provides a set of compatibility
    attributes for use with legacy code.

    Alias properties can optionally emit a deprecation warning on use, in order
    to help in the process of migrating legacy code.

    Version Changed:
        3.3:
        This now supports generics for typing, taking the types to return on
        access, and types that can be set.

    Example:
        .. code-block:: python

           class MyClass:
               new_prop: str
               old_prop: AliasProperty[int, str] = AliasProperty[int, str](
                   'new_prop',
                   convert_to_func=str,
                   convert_from_func=int)

        Note that the explicit type declaration is important. Without it,
        type checkers may allow constructors to override the type.
    """

    ######################
    # Instance variables #
    ######################

    #: Whether to emit a deprecation warning on access.
    deprecated: bool

    #: The type of class to use for the deprecation warning.
    deprecation_warning: Type[DeprecationWarning]

    #: The name of the property or attribute to read from and write to
    prop_name: str

    def __init__(
        self,
        prop_name: str,
        *,
        convert_to_func: Optional[Callable[[_AliasPropertySetT],
                                           _AliasPropertyStoredT]] = None,
        convert_from_func: Optional[Callable[[_AliasPropertyStoredT],
                                             _GetT]] = None,
        deprecated: bool = False,
        deprecation_warning: Type[DeprecationWarning] = DeprecationWarning,
    ) -> None:
        """Initialize the property.

        Version Changed:
            3.3:
            All arguments but ``prop_name`` must be provided as keyword
            arguments. This will be enforced in Djblets 5.0.

        Args:
            prop_name (str):
                The name of the property or attribute to read from and write
                to.

            convert_to_func (callable, optional):
                An optional function to call on a value before setting it on
                the aliased property name. This must take in the value as a
                parameter and return a value to set.

            convert_from_func (callable, optional):
                An optional function to call on a value after accessing it on
                the aliased property name and before returning to the caller.
                This must take in the value from the aliased property and
                return a value to return to the caller.

            deprecated (bool, optional):
                Whether to emit a deprecation warning when setting or accessing
                the property.

            deprecation_warning (type):
                The type of class to use for the deprecation warning. This
                should be a subclass of :py:class:`DeprecationWarning`.
        """
        self.prop_name = prop_name
        self.deprecated = deprecated
        self.deprecation_warning = deprecation_warning

        self._convert_to_func = convert_to_func
        self._convert_from_func = convert_from_func

    def __set__(
        self,
        instance: object,
        value: _AliasPropertySetT,
    ) -> None:
        """Set a value on the property.

        This will convert the value (if ``convert_to_func`` was provided
        to this property) and set it on the aliased property.

        If this is a deprecated property, this will emit a warning.

        Args:
            instance (object):
                The instance owning this property.

            value (object):
                The value being set.
        """
        if self.deprecated:
            cls_name = type(instance).__name__
            warnings.warn('%s.%s is deprecated. Please set %s.%s instead.'
                          % (cls_name, self.attr_name,
                             cls_name, self.prop_name),
                          self.deprecation_warning,
                          stacklevel=2)

        if value is not None and self._convert_to_func is not None:
            value = self._convert_to_func(value)

        setattr(instance, self.prop_name, value)

    @overload
    def __get__(
        self,
        instance: None,
        owner: Type[object],
    ) -> Self:
        ...

    @overload
    def __get__(
        self,
        instance: object,
        owner: Type[object],
    ) -> _GetT:
        ...

    def __get__(
        self,
        instance: object,
        owner: Type[object],
    ) -> Union[Self, _GetT]:
        """Return the value of the property.

        This will retrieve the value from the aliased property, converting
        it (if ``convert_from_func`` was provided to this property), and
        return it.

        If this is a deprecated property, this will emit a warning.

        Args:
            instance (object):
                The instance owning this property.

            owner (type):
                The instance's class.

        Returns:
            object:
            The property value.
        """
        if instance is None:
            return self

        if self.deprecated:
            cls_name = type(instance).__name__
            warnings.warn('%s.%s is deprecated. Please access %s.%s instead.'
                          % (cls_name, self.attr_name,
                             cls_name, self.prop_name),
                          self.deprecation_warning,
                          stacklevel=2)

        value = getattr(instance, self.prop_name)

        if value is not None and self._convert_from_func is not None:
            value = self._convert_from_func(value)

        return value


class TypedProperty(Generic[_TypedPropertyGetT, _SetT],
                    BaseProperty[_TypedPropertyGetT]):
    """A property that enforces type safety.

    This property will ensure that only values that are compatible with a
    given type can be set. This ensures type safety and helps catch errors
    early.

    Version Changed:
        3.3:
        This now supports generics for typing, taking the types to return on
        access, and types that can be set.

    Example:
        .. code-block:: python

           class MyClass:
               optional_prop: TypedProperty[Optional[str], Optional[str]] = \
                   TypedProperty(str)
               required_to_set_prop: TypedProperty[Optional[int], int] = \
                   TypedProperty(int,
                                 default=None
                                 allow_none=False)
               never_none_prop: TypedProperty[int, int] = \
                   TypedProperty(int,
                                 default=42,
                                 allow_none=False)

        Note that the explicit type declaration is important. Without it,
        type checkers may allow constructors to override the type.
    """

    ######################
    # Instance variables #
    ######################

    #: Whether a ``None`` value is allowed to be set.
    allow_none: bool

    #: The default value for the property if one is not set.
    default: Optional[_TypedPropertyGetT]

    #: The types that are valid for this property.
    #:
    #: New values are checked against this at runtime.
    valid_types: Tuple[Type[_SetT], ...]

    def __init__(
        self,
        valid_types: _TypedPropertyValidTypesParamT,
        *,
        default: Optional[_TypedPropertyGetT] = None,
        allow_none: bool = True,
    ) -> None:
        """Initialize the property.

        Version Changed:
            3.3:
            All arguments but ``prop_name`` must be provided as keyword
            arguments. This will be enforced in Djblets 5.0.

        Args:
            valid_types (list of type):
                The types of values that are permitted to be set.

            default (object, optional):
                The default value, if one is not set.

            allow_none (bool, optional):
                Whether ``None`` values are allowed to be set.
        """
        if isinstance(valid_types, tuple):
            self.valid_types = valid_types
        elif isinstance(valid_types, Sequence):
            self.valid_types = tuple(valid_types)
        else:
            self.valid_types = (valid_types,)

        self.default = default
        self.allow_none = allow_none

    def __set__(
        self,
        instance: object,
        value: _SetT,
    ) -> None:
        """Set a value on the property.

        This will check if the value is of a valid type, and then set it on
        the instance.

        Args:
            instance (object):
                The instance owning this property.

            value (object):
                The value being set.

        Raises:
            TypeError:
                The value is not of a valid type.
        """
        if value is None and not self.allow_none:
            raise TypeError('None is not a valid value for this property.')

        if value is not None and not isinstance(value, self.valid_types):
            raise TypeError('%s (%r) is not a valid type for this property.'
                            % (type(value).__name__, value))

        instance.__dict__['_%s_typed' % self.attr_name] = value

    @overload
    def __get__(
        self,
        instance: None,
        owner: Type[object],
    ) -> Self:
        ...

    @overload
    def __get__(
        self,
        instance: object,
        owner: Type[object],
    ) -> _TypedPropertyGetT:
        ...

    def __get__(
        self,
        instance: object,
        owner: Type[object],
    ) -> Union[Self, _TypedPropertyGetT]:
        """Return the value of the property.

        Args:
            instance (object):
                The instance owning this property.

            owner (type):
                The instance's class.

        Returns:
            object:
            The property value.
        """
        if instance is None:
            return self

        return cast(
            _TypedPropertyGetT,
            instance.__dict__.get('_%s_typed' % self.attr_name, self.default))
